from typing import Optional, List, Any, Dict, Union
from banditbench.sampling.sampler import DatasetBuffer
from banditbench.agents.llm import LLMMABAgentBase, LLMCBAgentBase

class FewShot:
    """
    Takes in agent, and returns a FewShot agent
    FewShot.empower(agent, *configs) # becomes a FewShotAgent

    Agent stores the state, not FewShot class

    This is a passthrough class.
    It augments the original agent's behavior with few-shot examples.

    We use class variables to help applying the same config to different agents without reprocessing
    """

    data_buffer: Optional[DatasetBuffer] = None
    filename: Optional[str] = None
    num_examples: int = 5
    skip_first: int = 2
    sample_freq: int = 2
    traj_idx: int = 0  # which trajectory we choose from the datasource

    @classmethod
    def empower(cls, agent: Union[LLMMABAgentBase, LLMCBAgentBase],
                datasource: Optional[Union[DatasetBuffer, str]] = None,
                num_examples: Optional[int] = None,
                skip_first: Optional[int] = None,
                sample_freq: Optional[int] = None,
                traj_idx: Optional[int] = None) -> Union[LLMMABAgentBase, LLMCBAgentBase]:
        """
        This is the general "compilation" function that takes in the agent

        :param skip_first: skip the first few examples (because the decision might not be very complex)
        :param num_examples: The total number of examples in context
        :param sample_freq: For each trajectory, the number of improvement steps are between each example
        """
        if datasource is not None:
            print("received datasource: ", datasource)
            if type(datasource) is str:
                cls.filename = datasource
                cls.data_buffer = DatasetBuffer.load(cls.filename)
            elif type(datasource) is DatasetBuffer:
                cls.data_buffer = datasource
            else:
                raise ValueError("Invalid datasource type")

        if traj_idx is not None:
            cls.traj_idx = traj_idx

        if num_examples is not None:
            cls.num_examples = num_examples

        if skip_first is not None:
            cls.skip_first = skip_first

        if sample_freq is not None:
            cls.sample_freq = sample_freq

        # then determine MAB or CB to load the examples into agent
        if isinstance(agent, LLMMABAgentBase):
            agent.demos = cls.load_few_shot_mab_examples(agent)
        elif isinstance(agent, LLMCBAgentBase):
            agent.demos = cls.load_few_shot_cb_examples(agent)

        return agent

    @classmethod
    def load_few_shot_mab_examples(cls, agent: Union[LLMMABAgentBase, LLMCBAgentBase]) -> Optional[str]:
        """This has to be different """
        if cls.data_buffer is None:
            return None
        else:
            fewshot_prompt = "\n" + agent.env.bandit_scenario.fewshot_prompt
            fewshot_prompt += "========================"
            start_idx = cls.skip_first
            # first trajectory in the data buffer
            examples = cls.data_buffer[cls.traj_idx].verbal_prompts[start_idx::cls.sample_freq][:cls.num_examples]
            if len(examples) == 0:
                return None
            for example in examples:
                fewshot_prompt += example['action_history']
                # query
                fewshot_prompt += example['decision_query']
                fewshot_prompt += f"\n{example['label']}\n"
                fewshot_prompt += "========================"

            return fewshot_prompt

    @classmethod
    def load_few_shot_cb_examples(cls, agent: Union[LLMMABAgentBase, LLMCBAgentBase]) -> Optional[str]:
        """This has to be different """
        if cls.data_buffer is None:
            return None
        else:
            fewshot_prompt = "\n" + agent.env.bandit_scenario.fewshot_prompt
            fewshot_prompt += "============Example Trajectory============"
            start_idx = cls.skip_first
            examples = cls.data_buffer[cls.traj_idx].verbal_prompts[start_idx::cls.sample_freq][:cls.num_examples]
            if len(examples) == 0:
                return None
            for example in examples:
                fewshot_prompt += example['action_history']
                # query
                fewshot_prompt += example['decision_query']
                fewshot_prompt += f"\n{example['label']}\n"
                fewshot_prompt += "=============Example Trajectory==========="

            return fewshot_prompt



from typing import Union, Dict, Any, List
from banditbench.tasks.env import Bandit, VerbalBandit
from banditbench.tasks.cb.env import State

from pydantic import BaseModel


class Agent:
    name: str

    def __init__(self, env: Union[Bandit, VerbalBandit]) -> None:
        self.env = env
        self.k_arms = env.num_arms

    def reset(self):
        # no action
        pass


class MABAgent(Agent):

    def act(self) -> int:
        """Same as performing a sampling step."""
        raise NotImplementedError

    def update(self, action: int, reward: float, info: Dict[str, Any]) -> None:
        """The action performs an update step based on the action it chose, and the reward it received."""
        raise NotImplementedError


class CBAgent(Agent):
    def act(self, state: State) -> int:
        """Same as performing a sampling step."""
        raise NotImplementedError

    def update(self, state: State, action: int, reward: float, info: Dict[str, Any]) -> None:
        """The action performs an update step based on the action it chose, and the reward it received."""
        raise NotImplementedError


class ActionInfoField(BaseModel):
    info_name: str  # such as "exploitation value"
    info_template: Union[str, None]  # Note, we only support non-key-value templates, like "value name = {:.2f}"
    value: Union[float, str]

    def __init__(self, info_name: str, value: Union[float, str], info_template: Union[str, None] = None):
        super().__init__(info_name=info_name, value=value, info_template=info_template)

    def __str__(self):
        if self.info_template is None:
            if isinstance(self.value, float):
                return f"{self.info_name} {self.value:.2f}"
            else:
                return f"{self.info_name} {self.value}"
        else:
            return self.info_template.format(self.value)

    def to_str(self, json_fmt=False):
        """
        :param json_fmt: if set True, will mimic JSON format in string like `{info_name: value}` explicitly in string
        :return:
        """
        if json_fmt and self.info_template is None:
            if isinstance(self.value, float):
                return "{" + f"\"{self.info_name}\": {self.value:.2f}" + "}"
            else:
                return "{" + f"\"{self.info_name}\": {self.value}" + "}"
        else:
            return str(self)

    def __add__(self, other: Union['ActionInfoField', 'ActionInfo']):
        if isinstance(other, ActionInfoField):
            return ActionInfo(action_info_fields=[self, other])
        elif isinstance(other, ActionInfo):
            return ActionInfo(action_info_fields=[self] + other.action_info_fields)
        else:
            raise ValueError(f"Unsupported type: {type(other)}")


class ActionInfo(BaseModel):
    # an action can have multiple fields (of information)
    action_info_fields: List[ActionInfoField]

    def __str__(self):
        return ", ".join([info.to_str() for info in self.action_info_fields])

    def to_str(self, json_fmt=False):
        if not json_fmt:
            return str(self)
        else:
            return ", ".join([info.to_str(json_fmt=True) for info in self.action_info_fields])

    def __len__(self):
        return len(self.action_infos)

    def __add__(self, other: Union['ActionInfo', 'ActionInfoField']):
        if isinstance(other, ActionInfoField):
            return ActionInfo(action_info_fields=self.action_info_fields + [other])
        elif isinstance(other, ActionInfo):
            return ActionInfo(action_info_fields=self.action_info_fields + other.action_info_fields)
        else:
            raise ValueError(f"Unsupported type: {type(other)}")

    def get_info_by_name(self, info_name: str) -> Union[ActionInfoField, None]:
        """Retrieve an ActionInfoField by its info_name."""
        for field in self.action_info_fields:
            if field.info_name == info_name:
                return field
        return None

    def get_value_by_name(self, info_name: str) -> Union[float, str, None]:
        """Retrieve just the value of an ActionInfoField by its info_name."""
        field = self.get_info_by_name(info_name)
        if field is not None:
            return field.value
        return None

    def __getitem__(self, key: Union[str, int]) -> Union[float, str, None, ActionInfoField]:
        """Allow both dictionary-style access by info_name and list-style access by index."""
        if isinstance(key, str):
            return self.get_value_by_name(key)
        elif isinstance(key, int):
            if 0 <= key < len(self.action_info_fields):
                return self.action_info_fields[key]
            raise IndexError("Index out of range")
        raise TypeError(f"Invalid key type: {type(key)}")



import os
from typing import Optional
from banditbench.utils import dedent
from banditbench.tasks.scenario import MABScenario

class ScenarioUtil:
    """
    Inheriting from this class unifies the subclass' default __init__ method
    """
    def __init__(self, num_actions: int,
                 seed: Optional[int] = None):
        super().__init__(num_actions=num_actions, action_names=self.action_names, action_unit=self.action_unit, 
                         base_description=self.base_description, detailed_description=self.detailed_description,
                         query_prompt=self.query_prompt, seed=seed)

class ButtonPushing(ScenarioUtil, MABScenario):
  name = 'btn'
  
  # 100 actions
  action_names = [
      "blue",
      "green",
      "red",
      "yellow",
      "orange",
      "purple",
      "pink",
      "brown",
      "black",
      "white",
      "gray",
      "cyan",
      "magenta",
      "maroon",
      "olive",
      "lime",
      "teal",
      "navy",
      "indigo",
      "violet",
      "gold",
      "silver",
      "bronze",
      "coral",
      "turquoise",
      "lavender",
      "tan",
      "beige",
      "ivory",
      "aqua",
      "azure",
      "crimson",
      "fuchsia",
      "khaki",
      "peach",
      "plum",
      "sienna",
      "mint",
      "rose",
      "ruby",
      "salmon",
      "scarlet",
      "amber",
      "emerald",
      "jade",
      "mauve",
      "ochre",
      "periwinkle",
      "sapphire",
      "topaz",
      "chartreuse",
      "burgundy",
      "mustard",
      "cinnamon",
      "cobalt",
      "mint green",
      "forest green",
      "sky blue",
      "lilac",
      "tangerine",
      "raspberry",
      "pumpkin",
      "blush",
      "eggplant",
      "seafoam",
      "spruce",
      "lemon",
      "denim",
      "flamingo",
      "sand",
      "apricot",
      "honey",
      "chestnut",
      "midnight blue",
      "moss green",
      "bordeaux",
      "lavender blush",
      "slate gray",
      "gunmetal",
      "mint cream",
      "dark salmon",
      "persimmon",
      "cranberry",
      "wheat",
      "bistre",
      "magenta haze",
      "sepia",
      "ultramarine",
      "lime green",
      "steel blue",
      "blush pink",
      "pearl",
      "mulberry",
      "antique white",
      "burnt orange",
      "coral pink",
      "ice blue",
      "bright yellow",
      "honeydew",
      "baby blue",
  ]

  action_unit = "button"
  base_description = dedent(
      """You are a bandit algorithm in a room with {} buttons labeled {}.
                Each button is associated with a Bernoulli distribution with a fixed but unknown mean; the means for the two buttons could be different.
                For either button, when you press it, you will get a reward that is sampled from the button's associated distribution. Your goal is to maximize the total reward.
                """
  )

  detailed_description = dedent(
      """You are a bandit algorithm in a room with {} buttons labeled {}.
                Each button is associated with a Bernoulli distribution with a fixed but unknown mean; the means for the two buttons could be different.
                For either button, when you press it, you will get a reward that is sampled from the button's associated distribution. Your goal is to maximize the total reward .

                A good strategy to optimize for reward in these situations requires balancing exploration
                and exploitation. You need to explore to try out all of the buttons and find those
                with high rewards, but you also have to exploit the information that you have to
                accumulate rewards.
                """
  )

class OnlineAds(ScenarioUtil, MABScenario):
  name = 'ads'  

  action_names = [
      "A",
      "B",
      "C",
      "D",
      "E",
      "F",
      "G",
      "H",
      "I",
      "J",
      "K",
      "L",
      "M",
      "N",
      "O",
      "P",
      "Q",
      "R",
      "S",
      "T",
      "U",
      "V",
      "W",
      "X",
      "Y",
      "Z",
      "AA",
      "AB",
      "AC",
      "AD",
      "AE",
      "AF",
      "AG",
      "AH",
      "AI",
      "AJ",
      "AK",
      "AL",
      "AM",
      "AN",
      "AO",
      "AP",
      "AQ",
      "AR",
      "AS",
      "AT",
      "AU",
      "AV",
      "AW",
      "AX",
      "AY",
      "AZ",
      "BA",
      "BB",
      "BC",
      "BD",
      "BE",
      "BF",
      "BG",
      "BH",
      "BI",
      "BJ",
      "BK",
      "BL",
      "BM",
      "BN",
      "BO",
      "BP",
      "BQ",
      "BR",
      "BS",
      "BT",
      "BU",
      "BV",
      "BW",
      "BX",
      "BY",
      "BZ",
      "CA",
      "CB",
      "CC",
      "CD",
      "CE",
      "CF",
      "CG",
      "CH",
      "CI",
      "CJ",
      "CK",
      "CL",
      "CM",
      "CN",
      "CO",
      "CP",
      "CQ",
      "CR",
      "CS",
      "CT",
      "CU",
      "CV",
      "CW",
      "CX",
      "CY",
      "CZ",
  ]
  action_unit = "advertisement"
  base_description = dedent(
      """You are recommendation engine powered by a bandit algorithm that chooses advertisements to display to
                users when they visit your webpage. There are {} advertisements you can choose from,
                named {}. When a user visits the webpage you can choose an advertisement
                to display and you will observe whether the user clicks on the ad or not. You model
                this by assuming that each advertisement has a certain click rate and users click on
                advertisements with their corresponding rates."""
  )

  detailed_description = dedent(
      """You are recommendation engine powered by a bandit algorithm that chooses advertisements to display to
                users when they visit your webpage. There are {} advertisements you can choose from,
                named {}. When a user visits the webpage you can choose an advertisement
                to display and you will observe whether the user clicks on the ad or not. You model
                this by assuming that each advertisement has a certain click rate and users click on
                advertisements with their corresponding rates.

                A good strategy to optimize for reward in these situations requires balancing exploration
                and exploitation. You need to explore to try out all of the advertisements and find those
                with high rewards, but you also have to exploit the information that you have to
                accumulate rewards.
                """
  )


class VideoWatching(ScenarioUtil, MABScenario):
  name = 'vid'
  action_names = [
      "A",
      "B",
      "C",
      "D",
      "E",
      "F",
      "G",
      "H",
      "I",
      "J",
      "K",
      "L",
      "M",
      "N",
      "O",
      "P",
      "Q",
      "R",
      "S",
      "T",
      "U",
      "V",
      "W",
      "X",
      "Y",
      "Z",
      "AA",
      "AB",
      "AC",
      "AD",
      "AE",
      "AF",
      "AG",
      "AH",
      "AI",
      "AJ",
      "AK",
      "AL",
      "AM",
      "AN",
      "AO",
      "AP",
      "AQ",
      "AR",
      "AS",
      "AT",
      "AU",
      "AV",
      "AW",
      "AX",
      "AY",
      "AZ",
      "BA",
      "BB",
      "BC",
      "BD",
      "BE",
      "BF",
      "BG",
      "BH",
      "BI",
      "BJ",
      "BK",
      "BL",
      "BM",
      "BN",
      "BO",
      "BP",
      "BQ",
      "BR",
      "BS",
      "BT",
      "BU",
      "BV",
      "BW",
      "BX",
      "BY",
      "BZ",
      "CA",
      "CB",
      "CC",
      "CD",
      "CE",
      "CF",
      "CG",
      "CH",
      "CI",
      "CJ",
      "CK",
      "CL",
      "CM",
      "CN",
      "CO",
      "CP",
      "CQ",
      "CR",
      "CS",
      "CT",
      "CU",
      "CV",
      "CW",
      "CX",
      "CY",
      "CZ",
  ]
  action_unit = "video"
  base_description = dedent(
      """You are a video recommendation system powered by a bandit algorithm for an online streaming platform. 
    There are {} videos available in your library, titled {}. 
    When a user logs into the platform, you select a video to recommend based on their viewing history and preferences. 
    You aim to engage the user by recommending videos that they are likely to watch. 
    Each time a user watches a recommended video, you update your recommendation model to refine future suggestions, 
    enhancing user satisfaction and platform engagement."""
  )

  detailed_description = dedent(
      """You are a video recommendation system powered by a bandit algorithm for an online streaming platform. 
    There are {} videos available in your library, titled {}. 
    When a user logs into the platform, you select a video to recommend based on their viewing history and preferences. 
    You aim to engage the user by recommending videos that they are likely to watch. 
    Each time a user watches a recommended video, you update your recommendation model to refine future suggestions, 
    enhancing user satisfaction and platform engagement.

    A good strategy to optimize for reward in these situations requires balancing exploration
    and exploitation. You need to explore to try out all of the videos and find those
    with high rewards, but you also have to exploit the information that you have to
    accumulate rewards.
    """
  )

class ClothesShopping(ScenarioUtil, MABScenario):
  name = 'clothes'
  
  action_names = [
      "Velvet Vogue Jacket",
      "Silk Serenity Dress",
      "Urban Mystique Jeans",
      "Celestial Symphony Scarf",
      "Retro Revival Sneakers",
      "Ethereal Elegance Blouse",
      "Midnight Mirage Trousers",
      "Vintage Vibe Coat",
      "Opulent Oasis Gown",
      "Mystic Mosaic Shirt",
      "Chic Cascade Skirt",
      "Radiant Reverie Blazer",
      "Bohemian Bliss Boots",
      "Stellar Style Suit",
      "Luminous Legend Jacket",
      "Panache Paradise Pants",
      "Gilded Galaxy Dress",
      "Posh Phoenix Pullover",
      "Enchanté Enigma Ensemble",
      "Luxe Loft Leggings",
      "Bespoke Bliss Blouse",
      "Dapper Dreams Denim",
      "Glamour Gaze Gown",
      "Eminent Essence Earrings",
      "Avant-Garde Allure Anklets",
      "Refined Radiance Romper",
      "Sophisticated Silhouette Sweater",
      "Opulent Oasis Overcoat",
      "Noble Nimbus Nightwear",
      "Divine Debonair Dungarees",
      "Regal Reverie Raincoat",
      "Marvelous Mirage Mittens",
      "Extravagant Expanse Earmuffs",
      "Ingenious Impressions Iridescent",
      "Prestige Paradigm Parka",
      "Dazzling Dynasty Dress",
      "Supreme Sylvan Sandals",
      "Elite Echoes Espadrilles",
      "Sublime Sashay Scarf",
      "Opulent Odyssey Overalls",
      "Elegant Eclipse Ensemble",
      "Refined Renaissance Robe",
      "Lavish Luminary Legwarmers",
      "Sculpted Serenity Shawl",
      "Timeless Titan Turtleneck",
      "Whimsical Whimsy Wrap",
      "Piquant Prestige Poncho",
      "Noble Nexus Nightshirt",
      "Faithful Fantasy Frock",
      "Radiant Realm Rainboots",
      "Alluring Apex Apron",
      "Delightful Dusk Dress",
      "Prestigious Prism Pants",
      "Gossamer Gleam Gloves",
      "Supreme Spectrum Slippers",
      "Celestial Chic Cape",
      "Magnetic Melody Mittens",
      "Glacial Grace Gaiters",
      "Gleaming Galore Gown",
      "Infinite Impeccable Jacket",
      "Viridian Visage Vest",
      "Ornamental Oasis Overalls",
      "Enchanted Eden Earmuffs",
      "Aesthetic Aura Anorak",
      "Mystical Muse Midi",
      "Exquisite Expanse Earrings",
      "Relished Realm Romper",
      "Glamorous Gala Gown",
      "Serenade Spirit Shawl",
      "Charmed Cascade Cardigan",
      "Splendid Solstice Sweater",
      "Majestic Manor Moccasins",
      "Titanic Tempest Tunic",
      "Royal Reverie Wrap",
      "Dazzling Delight Dress",
      "Verdant Vogue Vest",
      "Ethereal Expanse Earrings",
      "Sublime Symphony Sweater",
      "Pristine Plaid Pullover",
      "Opulent Oasis Outerwear",
      "Fabled Flora Frock",
      "Glittering Gala Gloves",
      "Eminent Envy Earwarmer",
      "Classic Couture Cardigan",
      "Mythic Mirage Muffler",
      "Silk Spectrum Slip",
      "Refined Rascal Rainjacket",
      "Splendid Sorcery Socks",
      "Lustrous Lattice Leggings",
      "Radiant Ripple Romper",
      "Stellar Sheen Shawl",
      "Serenade Spirit Suit",
      "Opal Obsession Overcoat",
      "Bejeweled Bloom Blazer",
      "Ornate Oratorio Onesie",
      "Dapper Dusk Duffle Coat",
      "Prismatic Paradise Poncho",
      "Majestic Moonlit Mantle",
      "Vivid Velocity Vestment",
      "Radiant Radiance Rucksack",
  ]

  action_unit = "item"
  base_description = dedent(
      """You are an AI fashion assistant for an online boutique powered by a bandit algorithm that offers a variety of clothing options from different brands. 
    There are {} unique clothing items you can recommend, named {}. 
    When a customer visits the online store, you assess their style preferences and shopping history to choose an item to suggest. 
    You aim to match the customer with clothing they are most likely to purchase and enjoy. 
    Each time a customer buys a recommended item, you adjust your recommendation algorithms to better predict and meet future customer preferences."""
  )

  detailed_description = dedent(
      """You are an AI fashion assistant for an online boutique powered by a bandit algorithm that offers a variety of clothing options from different brands. 
    There are {} unique clothing items you can recommend, named {}. 
    When a customer visits the online store, you assess their style preferences and shopping history to choose an item to suggest. 
    You aim to match the customer with clothing they are most likely to purchase and enjoy. 
    Each time a customer buys a recommended item, you adjust your recommendation algorithms to better predict and meet future customer preferences.

    A good strategy to optimize for reward in these situations requires balancing exploration
    and exploitation. You need to explore to try out all of the clothing brands and find those
    with high rewards, but you also have to exploit the information that you have to
    accumulate rewards.
    """
  )



import json
import numpy as np
from pathlib import Path
from typing import Union, Dict, Any, List
from banditbench.tasks.env import VerbalBandit, Bandit
from banditbench.tasks.cb.env import ContextualBandit
from banditbench.tasks.types import Trajectory
from banditbench.agents.types import Agent, ActionInfo

from banditbench.utils import plot_cumulative_reward, plot_multi_cumulative_reward

"""
DatasetBuffer has 3 components:
 - Trajectories: List of Trajectory objects (bare minimum) (all agents will have this) (this is just the raw interaction history with the environment)
 - ActionInfos: Additional information at each step of the decision (some agents have them, some don't) (for agent that has them, this is not exposed currently)
 - VerbalPrompts: The prompt, task description that was sent into LLM to get the label (For LLM agent, and oracleLLM agent) (these are also not exposed)
"""


class Data(dict):
    # this is on the trajectory level -- a single trajectory
    trajectory: Trajectory
    ag_info: Union[List[List[ActionInfo]], None]
    verbal_prompts: Union[List[Dict[str, str]], None]

    def __init__(self, trajectory: Trajectory, action_info: Union[List[List[ActionInfo]], None] = None,
                 verbal_prompts: Union[List[Dict[str, str]], None] = None):
        super().__init__()
        self['trajectory'] = trajectory
        self['action_info'] = action_info
        self['verbal_prompts'] = verbal_prompts

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(f"'Data' object has no attribute '{name}'")

    def __setattr__(self, name, value):
        self[name] = value

    def __iter__(self):
        return iter([self.trajectory, self.action_info, self.verbal_prompts])


class DatasetBuffer:
    # this is on the dataset level -- a dataset of trajectories
    def __init__(self, trajectories=None, ag_info=None, verbal_prompts=None):
        self.trajectories = trajectories or []
        self.ag_info = ag_info or []
        self.verbal_prompts = verbal_prompts or []

    def append(self, trajectory: Trajectory, action_info: Union[List[List[ActionInfo]], None] = None,
               verbal_prompt: Union[List[Dict[str, str]], None] = None):
        self.trajectories.append(trajectory)
        if action_info is not None:
            self.ag_info.append(action_info)
        if verbal_prompt is not None:
            self.verbal_prompts.append(verbal_prompt)

    def add(self, trajectory: Trajectory, action_info: Union[List[List[ActionInfo]], None] = None,
            verbal_prompt: Union[List[Dict[str, str]], None] = None):
        self.append(trajectory, action_info, verbal_prompt)

    def clear(self):
        self.trajectories.clear()
        self.ag_info.clear()
        self.verbal_prompts.clear()

    def __len__(self):
        return len(self.trajectories)

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            # Handle slice indexing
            trajectories = self.trajectories[idx]
            ag_info = self.ag_info[idx] if self.ag_info else None
            verbal_prompts = self.verbal_prompts[idx] if self.verbal_prompts else None

            # Create new buffer with sliced data
            new_buffer = DatasetBuffer(trajectories, ag_info, verbal_prompts)
            return new_buffer
        else:
            # Handle single index
            return Data(
                trajectory=self.trajectories[idx],
                action_info=self.ag_info[idx] if self.ag_info else None,
                verbal_prompts=self.verbal_prompts[idx] if self.verbal_prompts else None
            )

    def __str__(self):
        return f"DatasetBuffer({len(self)} trajectories)"

    def __repr__(self):
        return str(self)

    def __iter__(self):
        for i in range(len(self)):
            yield Data(
                trajectory=self.trajectories[i],
                action_info=self.ag_info[i] if self.ag_info else None,
                verbal_prompts=self.verbal_prompts[i] if self.verbal_prompts else None
            )

    def __add__(self, other):
        if isinstance(other, DatasetBuffer):
            result = DatasetBuffer()
            result.trajectories.extend(self.trajectories)
            result.trajectories.extend(other.trajectories)
            if self.ag_info and other.ag_info:
                result.ag_info.extend(self.ag_info)
                result.ag_info.extend(other.ag_info)
            if self.verbal_prompts and other.verbal_prompts:
                result.verbal_prompts.extend(self.verbal_prompts)
                result.verbal_prompts.extend(other.verbal_prompts)
            return result
        else:
            raise ValueError(f"Unsupported type: {type(other)}")

    def dump(self, file):
        """Save the dataset buffer to a JSON file."""
        if isinstance(file, str):
            filepath = file
        else:
            filepath = file.name

        data = {
            'n_trajectories': len(self),
            'trajectories': [
                traj.model_dump() for traj in self.trajectories
            ]
        }

        if self.ag_info:
            data['ag_info'] = [
                [[info.model_dump() for info in action_infos]
                 for action_infos in interaction_infos]
                for interaction_infos in self.ag_info
            ]

        if self.verbal_prompts:
            data['verbal_prompts'] = self.verbal_prompts

        with open(filepath, 'w') as f:
            json.dump(data, f)

    @classmethod
    def load(cls, filepath: str) -> 'DatasetBuffer':
        """Load a dataset buffer from a JSON file."""
        with open(filepath, 'r') as f:
            data = json.load(f)

        trajectories = [Trajectory.model_validate(traj_data) for traj_data in data['trajectories']]
        buffer = cls(trajectories=trajectories)

        if 'ag_info' in data and data['ag_info']:
            buffer.action_infos = [
                [
                    [ActionInfo.model_validate(info) for info in action_infos]
                    for action_infos in interaction_infos
                ]
                for interaction_infos in data['ag_info']
            ]

        if 'verbal_prompts' in data:
            buffer.verbal_prompts = data['verbal_prompts']

        return buffer

    def save(self, file):
        self.dump(file)

    def plot_performance(self, *args, title=None, labels=None, filename=None):
        # plot the mean performance over all trajectories stored in the dataset
        config_name_to_all_rewards = {}
        
        # Parse args to find title if provided
        buffers = []
        for arg in args:
            if isinstance(arg, str):
                if arg.endswith('.png') or arg.endswith('.jpg') or arg.endswith('.jpeg') or arg.endswith('.pdf'):
                    filename = arg
                else:
                    title = arg
            else:
                buffers.append(arg)
        
        # Get rewards from current buffer
        current_rewards = []
        for trajectory in self.trajectories:
            rewards = []
            for interaction in trajectory:
                rewards.append(interaction.reward)
            current_rewards.append(rewards)
            
        config_name = labels[0] if labels and len(labels) > 0 else 'Agent 1'
        config_name_to_all_rewards[config_name] = current_rewards
        
        # Get rewards from other buffers if provided
        if buffers:
            for i, buffer in enumerate(buffers):
                buffer_rewards = []
                for trajectory in buffer.trajectories:
                    rewards = []
                    for interaction in trajectory:
                        rewards.append(interaction.reward)
                    buffer_rewards.append(rewards)
                    
                buffer_name = labels[i+1] if labels and len(labels) > i+1 else f'Agent {i+2}'
                config_name_to_all_rewards[buffer_name] = buffer_rewards
            
        horizon = len(current_rewards[0])
        plot_multi_cumulative_reward(config_name_to_all_rewards, horizon, title, filename)

    def to_sft_data(self, file=None):
        """
        'task_instruction': task_instruction,
        'action_history': action_history,
        'decision_query': decision_query,
        'label': action_verbal
        """
        # [{}]
        data = []
        for trajectory_prompts in self.verbal_prompts:
            traj_prompt = []
            for i, trajectory_step_prompt in enumerate(trajectory_prompts):
                traj_prompt.append({'step': i,
                                    "prompt": trajectory_step_prompt['task_instruction'] + trajectory_step_prompt[
                                        'action_history'] + trajectory_step_prompt['decision_query'],
                                    "label": trajectory_step_prompt['label']})
            data.append(traj_prompt)

        if file:
            if isinstance(file, str):
                filepath = file
            else:
                filepath = file.name

            with open(filepath, 'w') as f:
                json.dump(data, f)
        else:
            return data

    def save_sft_data(self, file=None):
        return self.to_sft_data(file)



from banditbench.tasks.cb.movielens import MovieLens, MovieLensVerbal

def test_construct_env():
    env = MovieLens('100k-ratings', 5, 20, 5, 'train', save_data_dir='./tensorflow_datasets/')
    state, _ = env.reset()
    verbal_env = MovieLensVerbal(env)
    verbal_state, info = verbal_env.reset()

test_construct_env()


from banditbench.tasks.mab.env import VerbalMultiArmedBandit, BernoulliBandit


def test_verbal_bandit(verbose=False):
    core_bandit = BernoulliBandit(2, 10, [0.5, 0.2], 123)
    verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching")
    assert len(verbal_bandit.history) == 0
    assert len(verbal_bandit.core_bandit.history) == 0

    query, reward, done, info = verbal_bandit.step("button 1")
    if verbose:
        print(query)
        print(reward)
        print(done)
        print(info)

def test_verbal_bandit_random():
    core_bandit = BernoulliBandit(2, 10, [0.5, 0.2], 123)
    verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching")
    assert len(verbal_bandit.history) == 0
    assert len(verbal_bandit.core_bandit.history) == 0

    _, _, _, info = verbal_bandit.step("button 1")
    assert info["is_random"] == True

def test_verbal_bandit_non_random():
    core_bandit = BernoulliBandit(2, 10, [0.5, 0.2], 123)
    verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching")
    action = verbal_bandit.action_names[0]
    _, _, _, info = verbal_bandit.step(action)
    assert info["is_random"] == False

def test_verbal_bandit_seed():
    core_bandit = BernoulliBandit(4, 10, [0.5, 0.2, 0.2, 0.2], 222)
    print(core_bandit.arm_params)
    verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching", scenario_seed=333)

    assert len(verbal_bandit.history) == 0
    assert len(verbal_bandit.core_bandit.history) == 0

    action = verbal_bandit.action_names[0]
    print(verbal_bandit.action_names)
    _, _, _, info = verbal_bandit.step(action)
    assert info["is_random"] == False

def test_verbal_bandit_history(verbose=False):
    core_bandit = BernoulliBandit(2, 10, [0.5, 0.2], 123)
    verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching")
    action = verbal_bandit.action_names[0]
    for _ in range(10):
        verbal_bandit.step(action)
    if verbose:
        print(verbal_bandit.history)
        print(len(verbal_bandit.history))
        print(len(verbal_bandit.core_bandit.history))
    assert len(verbal_bandit.history) == 10

def test_fewshot_loading():
    raise NotImplementedError

test_verbal_bandit()
test_verbal_bandit_random()
test_verbal_bandit_non_random()
test_verbal_bandit_seed()
test_verbal_bandit_history(verbose=True)


from typing import Optional, List, Any, Dict

import litellm

from banditbench.agents.context import ModelContextLayer, RawContextLayerMAB, RawContextLayerCB, SummaryContextLayerMAB, \
    SummaryAlgoGuideLayerMAB, RawContextAlgoGuideLayerCB

from banditbench.agents.guides import UCBGuide, LinUCBGuide
from banditbench.tasks.types import State
from banditbench.tasks.env import VerbalBandit
from banditbench.sampling.sampler import AgentSampleBase

import banditbench.tasks.cb as cb
import banditbench.tasks.mab as mab

from banditbench.agents.types import MABAgent, CBAgent


class LLM:
    """Base class for LLM functionality shared across agent types."""

    def __init__(self, model: str = "gpt-3.5-turbo",
                 api_base: Optional[str] = None):
        """Initialize LLM agent with specified model.
        
        Args:
            model: Name of LLM model to use (default: gpt-3.5-turbo)
        """
        self.model = model

    def generate(self, message: str) -> str:
        """Generate LLM response for given messages.

        Returns:
            Generated response text
        """
        response = litellm.completion(
            model=self.model,
            messages=[{"content": message, "role": "user"}]
        )
        return response.choices[0].message.content


class LLMMABAgentBase(MABAgent, LLM, ModelContextLayer, AgentSampleBase):
    """LLM-based multi-armed bandit agent."""

    interaction_history: List[mab.VerbalInteraction]
    demos: Optional[str]

    decision_context_start: str = "So far you have interacted {} times with the following choices and rewards:\n"

    def __init__(self, env: VerbalBandit,
                 model: str = "gpt-3.5-turbo",
                 history_context_len=1000,
                 verbose=False):
        MABAgent.__init__(self, env)
        LLM.__init__(self, model)
        ModelContextLayer.__init__(self, history_context_len)
        self.demos = None  # few-shot demos, not reset, and only specified by FewShot class
        self.verbose = verbose

    def act(self) -> str:
        """Generate next action using LLM."""
        # Implement LLM-based action selection
        n_interactions = len(self.interaction_history)

        task_instruction = self.env.get_task_instruction()
        if self.demos is not None:
            task_instruction += self.demos + '\n'
        history_context = self.decision_context_start.format(n_interactions) + self.represent_history()
        query = self.env.get_query_prompt()

        response = self.generate(task_instruction + history_context + query)
        return response

    def update(self, action: int, reward: float, info: Dict[str, Any]) -> None:
        # note that we don't have access to the expected reward on the agent side
        assert 'interaction' in info
        assert type(info['interaction']) is mab.VerbalInteraction
        self.interaction_history.append(info['interaction'])

    def represent_history(self) -> str:
        return self.represent_interaction_context(self.env.action_names, self.env.bandit_scenario.action_unit,
                                                  self.history_context_len)

    def reset(self):
        super().reset()  # MABAgent.reset()
        self.interaction_history = []

    def get_task_instruction(self) -> str:
        task_instruction = self.env.get_task_instruction()
        return task_instruction

    def get_action_history(self) -> str:
        history_context = self.represent_history()
        return history_context

    def get_decision_query(self) -> str:
        query = self.env.get_query_prompt()
        return query


class LLMCBAgentBase(CBAgent, LLM, ModelContextLayer, AgentSampleBase):
    """LLM-based contextual bandit agent."""

    interaction_history: List[cb.VerbalInteraction]
    demos: Optional[str]  # few-shot demos, not reset, and only specified by FewShot class

    decision_context_start: str = ("So far you have interacted {} times with the most recent following choices and "
                                   "rewards:\n")

    def __init__(self, env: VerbalBandit,
                 model: str = "gpt-3.5-turbo",
                 history_context_len=1000,
                 verbose=False):
        CBAgent.__init__(self, env)
        LLM.__init__(self, model)
        ModelContextLayer.__init__(self, history_context_len)
        self.demos = None
        self.verbose = verbose

    def act(self, state: State) -> str:
        """Generate next action using LLM and context."""
        # Implement LLM-based contextual action selection
        n_interactions = len(self.interaction_history)

        task_instruction = self.env.get_task_instruction()
        if self.demos is not None:
            task_instruction += self.demos + '\n'
        history_context = self.decision_context_start.format(n_interactions) + self.represent_history()
        query = self.env.get_query_prompt(state, side_info=None)

        response = self.generate(task_instruction + history_context + query)
        return response

    def represent_history(self) -> str:
        return self.represent_interaction_context(self.env.action_names, self.env.bandit_scenario.action_unit,
                                                  self.history_context_len)

    def update(self, state: State, action: int, reward: float, info: Dict[str, Any]) -> None:
        assert 'interaction' in info
        assert type(info['interaction']) is cb.VerbalInteraction
        self.interaction_history.append(info['interaction'])

    def reset(self):
        super().reset()  # MABAgent.reset()
        self.interaction_history = []

    def get_task_instruction(self) -> str:
        task_instruction = self.env.get_task_instruction()
        return task_instruction

    def get_action_history(self) -> str:
        history_context = self.represent_history()
        return history_context

    def get_decision_query(self, state: State) -> str:
        query = self.env.get_query_prompt(state, side_info=None)
        return query


class OracleLLMMABAgent(LLMMABAgentBase):
    """Not a full agent"""

    def __init__(self, env: VerbalBandit, oracle_agent: MABAgent,
                 model: str = "gpt-3.5-turbo", history_context_len=1000, verbose=False):
        """
        The oracle_agent will take action
        """

        super().__init__(env, model, history_context_len, verbose)
        self.oracle_agent = oracle_agent
        self.verbose = verbose

    def act(self) -> str:
        action_idx = self.oracle_agent.act()
        return self.env.action_names[action_idx]

    def update(self, action: int, reward: float, info: Dict[str, Any]) -> None:
        # note that we don't have access to the expected reward on the agent side
        assert 'interaction' in info
        assert type(info['interaction']) is mab.VerbalInteraction
        self.interaction_history.append(info['interaction'])
        self.oracle_agent.update(action, reward, info)

    def reset(self):
        super().reset()
        self.oracle_agent.reset()


class OracleLLMCBAgent(LLMCBAgentBase):
    """Not a full agent"""

    def __init__(self, env: VerbalBandit, oracle_agent: CBAgent,
                 model: str = "gpt-3.5-turbo", history_context_len=1000, verbose=False):
        """
        The oracle_agent will take action
        """

        super().__init__(env, model, history_context_len, verbose)
        self.oracle_agent = oracle_agent
        self.verbose = verbose

    def act(self, state: State) -> str:
        action_idx = self.oracle_agent.act(state)
        return self.env.action_names[action_idx]

    def update(self, state: State, action: int, reward: float, info: Dict[str, Any]) -> None:
        assert 'interaction' in info
        assert type(info['interaction']) is cb.VerbalInteraction
        self.interaction_history.append(info['interaction'])
        self.oracle_agent.update(state, action, reward, info)

    def reset(self):
        super().reset()
        self.oracle_agent.reset()


class LLMMABAgentSH(LLMMABAgentBase, SummaryContextLayerMAB):
    # MAB SH Agent
    name = "MAB_SH_Agent"


class LLMMABAgentRH(LLMMABAgentBase, RawContextLayerMAB):
    # MAB RH Agent
    name = "MAB_RH_Agent"


class LLMCBAgentRH(LLMCBAgentBase, RawContextLayerCB):
    # CB RH Agent
    name = "CB_RH_Agent"


class OracleLLMMABAgentSH(OracleLLMMABAgent, SummaryContextLayerMAB):
    name = "Oracle_MAB_SH_Agent"


class OracleLLMMAbAgentRH(OracleLLMMABAgent, RawContextLayerMAB):
    name = "Oracle_MAB_RH_Agent"


class OracleLLMCBAgentRH(OracleLLMCBAgent, RawContextLayerCB):
    name = "Oracle_CB_RH_Agent"


class LLMMABAgentSHWithAG(LLMMABAgentBase, LLM, SummaryAlgoGuideLayerMAB):
    name = "MAB_SH_AG_Agent"

    def __init__(self, env: VerbalBandit,
                 ag: UCBGuide,
                 model: str = "gpt-3.5-turbo",
                 history_context_len=1000,
                 verbose=False):
        LLMMABAgentBase.__init__(self, env)
        LLM.__init__(self, model)
        SummaryAlgoGuideLayerMAB.__init__(self, ag,
                                          history_context_len)
        self.verbose = verbose

    def update(self, action: int, reward: float, info: Dict[str, Any]) -> None:
        # note that we don't have access to the expected reward on the agent side
        super().update(action, reward, info)
        self.update_info_history(self.ag.get_actions_guide_info())
        self.update_algorithm_guide(action, reward, info)

    def reset(self):
        super().reset()  # LLMMABAgent.reset()
        self.ag.agent.reset()


class LLMCBAgentRHWithAG(LLMCBAgentBase, LLM, RawContextAlgoGuideLayerCB):
    def __init__(self, env: VerbalBandit,
                 ag: LinUCBGuide,
                 model: str = "gpt-3.5-turbo",
                 history_context_len=1000,
                 verbose=False):
        LLMCBAgentBase.__init__(self, env)
        LLM.__init__(self, model)
        RawContextAlgoGuideLayerCB.__init__(self, ag,
                                            history_context_len)
        self.verbose = verbose

    def update(self, state: State, action: int, reward: float, info: Dict[str, Any]) -> None:
        # note that we don't have access to the expected reward on the agent side
        super().update(state, action, reward, info)
        # store side information; this is the information we used to make the decision (because algorithm guide has
        # not been updated yet)
        self.update_info_history(self.ag.get_state_actions_guide_info(state))
        self.update_algorithm_guide(state, action, reward, info)

    def reset(self):
        super().reset()  # LLMCBAgent.reset()
        self.ag_info_history = []
        self.ag.agent.reset()

    def act(self, state: State) -> str:
        """Generate next action using LLM and context."""
        # Implement LLM-based contextual action selection
        task_instruction = self.env.get_task_instruction()
        history_context = self.represent_interaction_context()

        ag_info = self.ag.get_state_actions_guide_info(state)
        snippet = ""
        for i, action_info in enumerate(ag_info):
            # normal format
            # snippet += '\n' + action_names[i].split(") (")[0] + ")" + ": " + action_info.to_str()

            # JSON-like format used in the paper
            snippet += '\n{\"' + self.env.action_names[i] + ": " + action_info.to_str(
                json_fmt=True) + "}"
        snippet += '\n'

        query = self.decision_context_start + self.env.get_query_prompt(state, side_info=snippet)

        response = self.generate(task_instruction + history_context + query)
        return response


class OracleLLMMABAgentSHWithAG(OracleLLMMABAgent, LLM, SummaryAlgoGuideLayerMAB):
    name = "Oracle_MAB_SH_AG_Agent"

    def __init__(self, env: VerbalBandit,
                 ag: UCBGuide,
                 oracle_agent: MABAgent,
                 model: str = "gpt-3.5-turbo",
                 history_context_len=1000,
                 verbose=False):
        OracleLLMMABAgent.__init__(self, env, oracle_agent, model, history_context_len, verbose)
        LLM.__init__(self, model)
        SummaryAlgoGuideLayerMAB.__init__(self, ag,
                                          history_context_len)

    def update(self, action: int, reward: float, info: Dict[str, Any]) -> None:
        # note that we don't have access to the expected reward on the agent side
        assert 'interaction' in info
        assert type(info['interaction']) is mab.VerbalInteraction
        self.interaction_history.append(info['interaction'])
        self.oracle_agent.update(action, reward, info)
        self.update_info_history(self.ag.get_actions_guide_info())
        self.update_algorithm_guide(action, reward, info)

    def reset(self):
        super().reset()
        self.oracle_agent.reset()
        self.ag.agent.reset()


class OracleLLMCBAgentRHWithAG(OracleLLMCBAgent, LLM, RawContextAlgoGuideLayerCB):
    name = "Oracle_CB_RH_AG_Agent"

    def __init__(self, env: VerbalBandit,
                 ag: LinUCBGuide,
                 oracle_agent: CBAgent,
                 model: str = "gpt-3.5-turbo",
                 history_context_len=1000,
                 verbose=False):
        OracleLLMCBAgent.__init__(self, env, oracle_agent, model, history_context_len, verbose)
        LLM.__init__(self, model)
        RawContextAlgoGuideLayerCB.__init__(self, ag,
                                            history_context_len)

    def update(self, state: State, action: int, reward: float, info: Dict[str, Any]) -> None:
        assert 'interaction' in info
        assert type(info['interaction']) is cb.VerbalInteraction
        self.interaction_history.append(info['interaction'])
        self.oracle_agent.update(state, action, reward, info)
        self.update_info_history(self.ag.get_state_actions_guide_info(state))
        self.update_algorithm_guide(state, action, reward, info)

    def reset(self):
        super().reset()
        self.oracle_agent.reset()
        self.ag_info_history = []
        self.ag.agent.reset()

    def get_decision_query(self, state: State) -> str:
        # this is the only thing that challenges with side_info, for the current step
        ag_info = self.ag.get_state_actions_guide_info(state)
        snippet = ""
        for i, action_info in enumerate(ag_info):
            # normal format
            # snippet += '\n' + action_names[i].split(") (")[0] + ")" + ": " + action_info.to_str()

            # JSON-like format used in the paper
            snippet += '\n{\"' + self.env.action_names[i] + ": " + action_info.to_str(
                json_fmt=True) + "}"
        snippet += '\n'

        query = self.env.get_query_prompt(state, side_info=snippet)
        return query


class LLMAgentBuilder:
    """This is only a partially initialized agent"""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def build_with_env(self, env):
        return LLMAgent.build_with_env(env, *self.args, **self.kwargs)

    def build(self, env):
        return LLMAgent.build_with_env(env, *self.args, **self.kwargs)


class LLMAgent:
    @classmethod
    def build_with_env(cls, env, *args, **kwargs):
        # Extract ag, oracle_agent, and summary from either args or kwargs
        # we disallow passing in summary flag as an argument (because we can only determine `bool` from `args`, which is too risky)

        ag = None
        oracle_agent = None

        remaining_args = []
        for arg in args:
            if hasattr(arg, 'get_action_guide_info'):  # Check if arg is algorithm guide
                ag = arg
            elif hasattr(arg, 'act') and hasattr(arg,
                                                 'update'):  # Check if arg is oracle agent; can be LLM or a classic agent
                oracle_agent = arg
            # elif isinstance(arg, bool):  # Check if arg is summary flag
            #     summary = arg
            else:
                remaining_args.append(arg)

        # Also check kwargs
        ag = ag or kwargs.pop('ag', None)
        oracle_agent = oracle_agent or kwargs.pop('oracle_agent', None)
        summary = kwargs.pop('summary', False)

        # Determine if environment is contextual bandit or multi-armed bandit
        if hasattr(env, 'action_names'):
            is_cb = hasattr(env.core_bandit, 'sample_state')
        else:
            is_cb = hasattr(env, 'sample_state')

        if oracle_agent:
            if is_cb:
                if ag:
                    # When AG is present, always use SH
                    return OracleLLMCBAgentRHWithAG(env, ag, oracle_agent, *remaining_args, **kwargs)
                return OracleLLMCBAgentRH(env, oracle_agent, *remaining_args, **kwargs)
            else:  # MAB
                if ag:
                    # When AG is present, always use SH
                    return OracleLLMMABAgentSHWithAG(env, ag, oracle_agent, *remaining_args, **kwargs)
                # For MAB without AG, respect summary flag
                if summary:
                    return OracleLLMMABAgentSH(env, oracle_agent, *remaining_args, **kwargs)
                return OracleLLMMAbAgentRH(env, oracle_agent, *remaining_args, **kwargs)
        else:
            if is_cb:
                if ag:
                    # When AG is present, always use SH
                    return LLMCBAgentRHWithAG(env, ag, *remaining_args, **kwargs)
                return LLMCBAgentRH(env, *remaining_args, **kwargs)
            else:  # MAB
                if ag:
                    # When AG is present, always use SH
                    return LLMMABAgentSHWithAG(env, ag, *remaining_args, **kwargs)
                # For MAB without AG, respect summary flag
                if summary:
                    return LLMMABAgentSH(env, *remaining_args, **kwargs)
                return LLMMABAgentRH(env, *remaining_args, **kwargs)

    @classmethod
    def build(cls, *args, **kwargs):
        # Extract ag, oracle_agent, and summary from either args or kwargs
        # we disallow passing in summary flag as an argument (because we can only determine `bool` from `args`, which is too risky)

        ag = None
        oracle_agent = None
        env = None

        remaining_args = []
        for arg in args:
            if hasattr(arg, 'get_action_guide_info'):  # Check if arg is algorithm guide
                ag = arg
            elif hasattr(arg, 'act') and hasattr(arg,
                                                 'update'):  # Check if arg is oracle agent; can be LLM or a classic agent
                oracle_agent = arg
            elif hasattr(args, 'action_names'):
                env = arg
            else:
                remaining_args.append(arg)

        # Also check kwargs
        ag = ag or kwargs.pop('ag', None)
        oracle_agent = oracle_agent or kwargs.pop('oracle_agent', None)
        summary = kwargs.pop('summary', False)
        env = kwargs.pop('env', None)

        if env is not None:
            return cls.build_with_env(env, ag, oracle_agent, summary, *remaining_args, **kwargs)
        else:
            return LLMAgentBuilder(ag, oracle_agent, summary=summary)



import os
import re
import csv
import numpy as np

from typing import Dict, Any, Tuple, Union, List, Optional

from banditbench.tasks.types import Action, State, Info
from banditbench.tasks.env import VerbalBandit
from banditbench.tasks.cb.env import Interaction, VerbalInteraction

from banditbench.tasks.cb.env import ContextualBandit
from banditbench.tasks.cb.movielens.processing import load_data_files, load_movielens_data, movie_genre_to_text, \
    parse_int_list, safe_decode, get_geo_data
from banditbench.tasks.cb.movielens.scenario import MovieLensScenario


# Write it all here, but then break into
class MovieLens(ContextualBandit):
    def __init__(self,
                 task_name: str,
                 num_arms: int,
                 horizon: int,
                 rank_k: int = 10,
                 mode: str = 'eval',
                 seed: Optional[int] = None,
                 save_data_dir: Optional[str] = None) -> None:
        """
        task_name: str: the name of the dataset we load from the movielens directory
        rank_k: int: the dimension of the bandit arm context and user embedding. `rank_k` is only effective if it's smaller than num_arms.
                     rank_k = min(rank_k, num_arms)
        mode: str: 'eval' or 'train'. If 'train', we will load the training set, otherwise we load the test set.
        save_data_dir: point to the directory where the data is saved. If None, we default to Tensorflow Dataset.
        """

        assert mode in ['eval', 'train']
        assert '100k' in task_name or '1m' in task_name
        assert type(seed) == int or seed is None

        self.task_name = task_name
        self.num_arms = num_arms
        self.horizon = horizon
        self.rank_k = rank_k
        self.mode = mode
        self.set_seed(seed)

        self.save_data_dir = save_data_dir

        self.history = []
        self.h = 0

        if '100k' in self.task_name:
            self.start_user_idx = 0
            self.end_user_idx = 100
        elif '1m' in self.task_name:
            self.start_user_idx = 0
            self.end_user_idx = 200

        self.initialize_defaults()

        self._verbal_info = {
            'user_idx_to_feats': self._user_idx_to_feats,
            'movie_idx_to_feats': self._movie_idx_to_feats,
            'movie_genre_to_text': movie_genre_to_text
        }

    @property
    def feature_dim(self) -> int:
        return min(self.rank_k, self.num_arms)

    def initialize_defaults(self) -> None:
        """Note this is not a reset function. It loads necessary files and processes them."""

        # now we run SVD for matrix completion
        # We shuffle the context/users

        ratings_df = load_data_files(self.task_name, self.save_data_dir)

        self._data_matrix, self._user_idx_to_feats, self._movie_idx_to_feats = (
            load_movielens_data(ratings_df, top_k_movies=self.num_arms, seed=self.seed)
        )

        self._num_users, self._num_movies = self._data_matrix.shape
        self.n_actions = self._num_movies

        # Compute the SVD.
        u, s, vh = np.linalg.svd(self._data_matrix, full_matrices=False)

        # Keep only the largest singular values.
        self._u_hat = u[:, :self.rank_k].astype(np.float32)
        self._s_hat = s[:self.rank_k].astype(np.float32)
        self._v_hat = np.transpose(vh[:self.rank_k]).astype(np.float32)

        self._approx_ratings_matrix = np.matmul(
            self._u_hat * self._s_hat, np.transpose(self._v_hat)
        )

    @property
    def name(self) -> str:
        # cb_1m-ratings_arms10
        return f"cb_{self.task_name}_arms{self.num_arms}_mode_{self.mode}"

    @property
    def verbal_info(self) -> Dict[str, Any]:
        """
        CB might be able to provide additional information from the dataset about the state
        :return:
        """
        return self._verbal_info

    def reward_fn(self, state: State, action: Action) -> float:
        """In a contextual bandit, this is a function f(x, a)"""
        user_index = state.index
        return self._approx_ratings_matrix[user_index, action]

    def sample_state(self) -> State:
        """Returns the context for a given user.

        Returns:
            tuple: (svd_features, feat_dict)
        """
        if self.mode == 'train':
            user_index = self.np_random.integers(self.start_user_idx, self.end_user_idx).item()
        else:
            user_index = self.np_random.integers(self.end_user_idx, self._num_users).item()

        # Get user features from SVD
        svd_features = self._u_hat[user_index]
        user_features = self._user_idx_to_feats[user_index]

        # this user_feature is a partial information feature
        # the svd user feature is the full information feature
        feat_dict = {
            'user_features': user_features
        }

        state = State(feature=svd_features, feature_text=None, index=user_index, info=feat_dict)

        return state

    def step(self, state: State, action: Action):
        """Core step function that takes integer action and returns vector observation"""
        assert type(action) == int, "For contextual bandit, the action space is integer"

        self.h += 1
        if self.h < self.horizon:
            done = False
        else:
            done = True

        reward = self.reward_fn(state, action)
        self.actions_taken.append(action)
        self.avg_rewards.append(reward)

        self.history.append(Interaction(state, action, reward))

        obs = self.sample_state()

        return obs, reward, done, {}

    def reset(self, seed: Optional[int] = None) -> Tuple[State, Info]:
        self.avg_rewards = []
        self.actions_taken = []
        self.history = []
        self.h = 0

        if seed is not None:
            # ctx_seed decides sampling order
            self.set_seed(seed)

        obs = self.sample_state()

        return obs, None


class MovieLensVerbal(VerbalBandit):
    def __init__(self, core_bandit: MovieLens, instruction_type: str = "detailed") -> None:
        super().__init__(core_bandit)
        self.history = []
        self.core_bandit = core_bandit
        self.instruction_type = instruction_type
        self.bandit_scenario = MovieLensScenario(action_names=self.get_actions_text(),
                                                 num_actions=self.core_bandit.num_arms,
                                                 seed=None)  # no action shuffling
        self.initialize_defaults()

    def initialize_defaults(self) -> None:
        # https://github.com/scpike/us-state-county-zip/tree/master
        # load in csv "geo-data.csv" data from the data directory
        
        geo_data_path = get_geo_data()
        
        # Load the data
        self.zipdata = {}
        with open(geo_data_path) as f:
            csvfile = csv.DictReader(f)
            for row in csvfile:
                self.zipdata[row['zipcode']] = ("{} of {}".format(row['city'], row['county']), row['state'])

    @property
    def action_names(self) -> List[str]:
        # we actually remove genre information when we provide them to the user
        # .split(") (")[0] + ")\""
        return self.get_actions_text(genre=False)

    @property
    def name(self) -> str:
        return self.core_bandit.name  # There is only one scenario so no need to mark a difference

    def step(self, state: State, action: Action) -> Tuple[State, float, bool, Dict[str, Any]]:
        """
        :param action:   We expect the action string to follow this format:
                         `MovieTitle (Year)` or `MovieTitle (Year) (Genre1|Genre2|...)`
                         We do not take only the movie title out of the possibility of two movies with the same title
                         being released in two different years.
        :return:
        """

        # VerbalBandit takes in an action that's a string, which is the action name
        assert type(action) == str, "Action must be a string for VerbalBandit"

        action_id = self.action_parsing(action)
        is_random = False if action_id is not None else True
        if action_id is None:
            action_id = int(self.core_bandit.np_random.integers(0, self.core_bandit.num_arms))

        obs, reward, done, info = self.core_bandit.step(state, action_id)
        text = self.verbalize_state(obs)
        obs.feature_text = text

        feedback = self.verbalize_feedback(self.action_names[action_id], reward)

        interaction = VerbalInteraction(state, action, action_id,
                                        self.action_names[action_id],
                                        self.core_bandit.reward_fn(state, action_id),
                                        feedback,
                                        is_random)
        self.history.append(interaction)

        info['interaction'] = interaction
        info['is_random'] = is_random

        return obs, reward, done, info

    def verbalize_state(self, observation: State) -> str:
        return self.get_user_feat_text(observation.feature, observation.info['user_features'])

    def verbalize_feedback(self, action_name: Action, reward: float) -> str:
        assert type(action_name) == str, "Action name must be a string"
        feedback = "{} {}, reward {:.3f}".format(action_name, self.bandit_scenario.action_unit,
                                                 reward)
        return feedback

    def action_parsing(self, model_completion: str) -> Union[int, None]:
        """
        :param model_completion: We expect the completion string to follow this format:
                                 `MovieTitle (Year)` or `MovieTitle (Year) (Genre1|Genre2|...)`
                                 We do not take only the movie title out of the possibility of two movies with the same title
                                 being released in two different years.
        :return:
        """
        model_completion = model_completion.strip()
        query_parts = model_completion.split("(")
        try:
            title = query_parts[0].strip()
            year = query_parts[1].split(")")[0].strip()

            # Create a regex pattern to match the query
            pattern = rf"^{re.escape(title)}\s*\({re.escape(year)}\)"

            movie_list = self.bandit_scenario.action_names[: self.num_arms]
            for index, movie in enumerate(movie_list):
                if re.match(pattern, movie):
                    return index
        except:
            return None

    def reset(self, seed: Optional[int] = None) -> Tuple[State, Info]:
        self.history = []
        obs, _ = self.core_bandit.reset(seed)
        text = self.get_user_feat_text(obs.feature, obs.info['user_features'])
        obs.feature_text = text

        verbal_instruction = self.bandit_scenario.get_instruction(self.instruction_type)

        return obs, {'instruction': verbal_instruction, 'query': self.get_query_prompt(obs)}

    def get_query_prompt(self, state: State, side_info: Optional[str] = None) -> str:
        return self.bandit_scenario.get_query_prompt(state, side_info)

    def get_task_instruction(self, *args, **kwargs) -> str:
        verbal_instruction = self.bandit_scenario.get_instruction(self.instruction_type)
        return verbal_instruction

    def get_actions_text(self, genre=True):
        """
        this function returns "options" for the scenario:
        "Movie_Name (year) (genre1|genre2|...)"
        """

        options = []
        movie_idx_to_feats = self.core_bandit.verbal_info['movie_idx_to_feats']
        movie_genre_to_text = self.core_bandit.verbal_info['movie_genre_to_text']

        for i in range(len(movie_idx_to_feats)):
            movie_name, genres = movie_idx_to_feats[i]
            # avoid "Empire Strikes Back, The"
            # should be "The Empire Strikes Back"

            if isinstance(movie_name, bytes):
                movie_name = movie_name.decode()

            # get rid of encoding issues
            if "b'" == movie_name[:2] and "'" == movie_name[-1]:
                movie_name = movie_name[2:-1]

            if ', The' in movie_name:
                movie_name = str('The ' + movie_name.replace(', The', ''))

            genre_text = ''
            if type(genres) == str:
                genres = parse_int_list(genres)
            else:
                genres = genres.tolist()
            for g in genres:
                genre_text += movie_genre_to_text[int(g)] + '|'

            if genre:
                options.append(f'{movie_name} ({genre_text[:-1]})')
            else:
                options.append(f'{movie_name}')

        return options

    def generate_occupation_sentence(self, occupation, age, age_bucket, gender):
        occupation = (
            occupation.decode('utf-8')
            if isinstance(occupation, bytes)
            else occupation
        )

        if "b'" == occupation[:2] and "'" == occupation[-1]:
            occupation = occupation[2:-1]

        if age == -1:
            age = age_bucket

        # Define occupation categories and their corresponding templates
        employment_occupations = [
            'doctor',
            'technician',
            'artist',
            'writer',
            'engineer',
            'administrator',
            'librarian',
            'scientist',
            'programmer',
            'educator',
            'executive',
            'salesman',
            'lawyer',
        ]

        special_cases = {
            'student': (
                'This person is a {age}-year-old {gender} studying as a student'
            ),
            'healthcare': (
                'This person is a {age}-year-old {gender} working in the healthcare'
                ' field'
            ),
            'entertainment': (
                'This person is a {age}-year-old {gender} working in the'
                ' entertainment field'
            ),
            'marketing': (
                'This person is a {age}-year-old {gender} working in the marketing'
                ' field'
            ),
            'retired': 'This person is a {age}-year-old {gender} who is retired.',
            'homemaker': (
                'This person is a {age}-year-old {gender} working as a homemaker'
            ),
            'other': (
                'This person is a {age}-year-old {gender} with an occupation'
                " classified as 'other'"
            ),
            'none': (
                'This person is a {age}-year-old {gender} with no current'
                ' occupation'
            ),
        }

        if occupation in employment_occupations:
            return (
                f'This person is a {age}-year-old {gender}, working as a {occupation}'
            )
        elif occupation in special_cases:
            return special_cases[occupation].format(age=age, gender=gender)
        else:
            return (
                f'This person is a {age}-year-old {gender} with an occupation of'
                f' {occupation}'
            )

    def get_user_feat_text(self, obs, user_features):
        user_description_zipcode_template = (
            """ and live in an area with the zip code {zip_code}."""
        )
        user_description_full = """ and live in {county} county, {state}."""

        feature_text = (
            ' The user has some numerical values that represent'
            ' their true implicit preference or taste for all movies: {}.'
        )

        def get_gender(is_male):
            return 'man' if is_male else 'woman'

        def format_description(data):
            age_bucket, raw_age, is_male, _, occupation, zip_code = data
            age_gender_occupation_text = self.generate_occupation_sentence(
                occupation, int(raw_age), int(age_bucket), get_gender(is_male)
            )

            if safe_decode(zip_code) in self.zipdata:
                county, state = self.zipdata[safe_decode(zip_code)]
                return (
                        age_gender_occupation_text
                        + user_description_full.format(
                    gender=get_gender(is_male),
                    age=int(raw_age),
                    age_bucket=int(age_bucket),
                    occupation=safe_decode(occupation),
                    county=county,
                    state=state,
                )
                        + feature_text.format(str([n for n in obs.tolist()]))
                )
            else:
                return (
                        age_gender_occupation_text
                        + user_description_zipcode_template.format(
                    gender=get_gender(is_male),
                    age=int(raw_age),
                    age_bucket=int(age_bucket),
                    occupation=safe_decode(occupation),
                    zip_code=safe_decode(zip_code),
                )
                        + feature_text.format(str([n for n in obs.tolist()]))
                )

        description = format_description(user_features)
        return description



from typing import Dict, Any, Tuple, Union, List, Optional
from pydantic import BaseModel, field_serializer
import numpy as np
import json

Action = Union[int, str]
ExpectedReward = Union[float, None]
Info = Union[Dict[str, Any], None]


def safe_json_encode(obj):
    try:
        return json.dumps(obj)
    except:
        return None


class State(BaseModel):
    feature: Any  # must be numpy array
    feature_text: Union[str, None] = None  # a text representation of the feature
    index: Union[int, None]  # a pointer to the dataset (if there is a dataset)
    info: Info = None  # additional information

    @field_serializer('info', mode='plain')
    def serialize_info(self, info: Info, _info):
        """
        When info cannot be serialized, we return None to avoid triggering error
        """
        return safe_json_encode(info)

    @field_serializer('feature', mode='plain')
    def serialize_feature(self, feature: Any, _feature):
        """
        We perform an automatic numpy serialization
        """
        if type(feature) == np.ndarray:
            return feature.tolist()
        elif type(feature).__module__ == np.__name__:
            # this is a Numpy integer or float, i.e., np.int32
            return feature.item()
        return feature


class Trajectory(list):
    def __init__(self, interactions: Union[List['InteractionBase'], None] = None) -> None:
        super().__init__(interactions or [])

    def __add__(self, other: Union['InteractionBase', 'Trajectory']):
        if isinstance(other, InteractionBase):
            return Trajectory(list(self) + [other])
        elif isinstance(other, Trajectory):
            return Trajectory(list(self) + list(other))
        else:
            raise ValueError(f"Unsupported type: {type(other)}")

    def __getstate__(self):
        return list(self)

    def __setstate__(self, state):
        super().__init__(state)

    def __repr__(self) -> str:
        return f"Trajectory({super().__repr__()})"

    def model_dump(self, **kwargs) -> List[Dict[str, Any]]:
        return [
            item.model_dump(**kwargs) if hasattr(item, 'model_dump')
            else item.__dict__
            for item in self
        ]

    def model_dump_json(self, **kwargs) -> str:
        return json.dumps(self.model_dump(**kwargs))
    
    @classmethod
    def model_validate(cls, data: List[Dict[str, Any]]) -> 'Trajectory':
        """
        Create a Trajectory from a list of dictionaries containing interaction data.
        Similar to Pydantic's model_validate method.
        """

        # we do a local import to avoid looped import
        from banditbench.tasks.mab.env import Interaction as MABInteraction, VerbalInteraction as MABVerbalInteraction
        from banditbench.tasks.cb.env import Interaction as CBInteraction, VerbalInteraction as CBVerbalInteraction

        interactions = []
        for item in data:
            # Determine interaction type from the data structure
            if 'mapped_action_name' in item:
                # Verbal interaction
                if 'state' in item:
                    # CB verbal interaction has state
                    interaction = CBVerbalInteraction.model_validate(item)
                else:
                    # MAB verbal interaction has no state
                    interaction = MABVerbalInteraction.model_validate(item)
            else:
                # Regular interaction
                if 'state' in item:
                    # CB interaction has state
                    interaction = CBInteraction.model_validate(item)
                else:
                    # MAB interaction has no state
                    interaction = MABInteraction.model_validate(item)
            interactions.append(interaction)
            
        return cls(interactions)


class InteractionBase:
    def __add__(self, other: Union['InteractionBase', 'Trajectory']) -> Trajectory:
        if isinstance(other, InteractionBase):
            return Trajectory(interactions=[self, other])
        elif isinstance(other, Trajectory):
            return Trajectory(interactions=[self] + other.interactions)
        else:
            raise ValueError(f"Unsupported type: {type(other)}")



"""
Hosts sampler mixin (used by agent to add create dataset functionality)
"""
import json
import numpy as np
from pathlib import Path
from typing import Union, Dict, Any, List
from banditbench.tasks.env import VerbalBandit, Bandit
from banditbench.tasks.cb.env import ContextualBandit
from banditbench.tasks.types import Trajectory

from banditbench.sampling.types import DatasetBuffer

import concurrent.futures
from tqdm import tqdm

from copy import deepcopy

def get_trajectory_seeds(env_seed, n_trajs):
    """Get seeds for environment and trajectories.
    
    Args:
        env: The environment to get seeds for
        n_trajs: Number of trajectory seeds to generate
    Returns:
        List of trajectory seeds
    """
    # Get environment seed or generate random one
    if env_seed is None:
        env_seed = np.random.randint(0, 2**32-1)
        
    # Generate trajectory seeds
    rng = np.random.RandomState(env_seed)
    return rng.randint(0, 2**32-1, size=n_trajs)

class SampleBase:

    def in_context_learn(self, env: Union[Bandit, ContextualBandit], n_trajs=20, *args, **kwargs) -> DatasetBuffer:
        """Collect interactions from environment and store in buffer.
        
        Args:
            env: The environment to collect from (Verbal or non-verbal)
            agent: Agent to collect data with
            n_trajs: Number of self-improving trajectories to collect
        """
        # Check if environment is verbal by looking for verbal_info property
        is_verbal = hasattr(env, 'action_names')
        is_contextual = hasattr(env, 'feature_dim')

        traj_seeds = get_trajectory_seeds(env.seed, n_trajs)
        buffer = DatasetBuffer()

        for traj_seed in tqdm(traj_seeds, desc="Collecting trajectories"):
            trajectory = []
            self.reset()

            if is_contextual:
                # Contextual bandit case
                state, _ = env.reset(seed=traj_seed)
                done = False
                while not done:
                    action = self.act(state)
                    new_state, reward, done, info = env.step(state, action)
                    trajectory.append(info['interaction'])
                    self.update(state, action, reward, info)
                    state = new_state
            else:
                # Multi-armed bandit case
                env.reset(seed=traj_seed)
                done = False
                while not done:
                    action = self.act()
                    _, reward, done, info = env.step(action)
                    trajectory.append(info['interaction'])
                    self.update(action, reward, info)

            buffer.append(Trajectory(trajectory))

        return buffer
class VerbalGuideSampleBase:
    # Using AG to collect data will produce trajectory AND fill in side-info for each action

    def in_context_learn(self, env: Union[Bandit, ContextualBandit], n_trajs=20, *args, **kwargs) -> DatasetBuffer:
        # AG has an underlying agent
        # but also provides utility class to load in action info
        # we need to both get the interaction from the underlying agent
        # and collect the action info from the AG
        is_contextual = hasattr(env, 'feature_dim')

        traj_seeds = get_trajectory_seeds(env.seed, n_trajs)
        buffer = DatasetBuffer()

        for traj_seed in tqdm(traj_seeds, desc="Collecting trajectories"):
            trajectory = []
            ag_info = []

            self.agent.reset()

            if is_contextual:
                # Contextual bandit case
                state, _ = env.reset(seed=traj_seed)
                done = False
                while not done:
                    action = self.agent.act(state)
                    new_state, reward, done, info = env.step(state, action)
                    action_info = self.get_state_actions_guide_info(state)

                    trajectory.append(info['interaction'])
                    ag_info.append(action_info)

                    self.agent.update(state, action, reward, info)
                    state = new_state
            else:
                # Multi-armed bandit case
                env.reset(seed=traj_seed)
                done = False
                while not done:
                    action = self.agent.act()
                    _, reward, done, info = env.step(action)
                    action_info = self.get_actions_guide_info()

                    trajectory.append(info['interaction'])
                    ag_info.append(action_info)

                    self.agent.update(action, reward, info)

            buffer.add(Trajectory(trajectory), ag_info)

        return buffer

class AgentSampleBase:
    # Using LLMAgent to collect data will produce trajectory, fill in side-info for each action (optional), AND fill in verbal prompt
    # will fill in side-info only if `ag` is in the LLM Agent

    def in_context_learn(self, env: VerbalBandit, n_trajs=20, num_threads=10) -> DatasetBuffer:
        
        is_contextual = hasattr(env.core_bandit, 'feature_dim')
        buffer = DatasetBuffer()
        
        # Get trajectory seeds upfront
        traj_seeds = get_trajectory_seeds(env.core_bandit.seed, n_trajs)

        def collect_single_trajectory(trial_idx, trial_seed, trial_pbar):
            trajectory = []
            ag_info = []
            verbal_prompts = []
            
            # Create a new instance for thread safety
            env_copy = deepcopy(env)
            agent_copy = deepcopy(self)
            agent_copy.reset()

            # Set position to ensure proper display of nested progress bars
            step_pbar = tqdm(total=env_copy.horizon, 
                            desc=f'Trial {trial_idx+1} steps', 
                            leave=False,
                            position=trial_idx + 1)

            if is_contextual:
                # Contextual bandit case
                state, _ = env_copy.reset(seed=trial_seed)
                done = False
                while not done:
                    # Get verbal prompts for this step
                    task_instruction = agent_copy.get_task_instruction()
                    action_history = agent_copy.get_action_history()
                    decision_query = agent_copy.get_decision_query(state)

                    action_verbal = agent_copy.act(state)
                    verbal_prompts.append({
                        'task_instruction': task_instruction,
                        'action_history': action_history,
                        'decision_query': decision_query,
                        'label': action_verbal
                    })
                    new_state, reward, done, info = env_copy.step(state, action_verbal)
                    if hasattr(agent_copy, 'ag'):
                        action_info = agent_copy.ag.get_state_actions_guide_info(state)
                        ag_info.append(action_info)

                    trajectory.append(info['interaction'])
                    action = info['interaction'].mapped_action

                    agent_copy.update(state, action, reward, info)
                    state = new_state
                    step_pbar.update(1)
            else:
                # Multi-armed bandit case  
                env_copy.reset(seed=trial_seed)
                done = False
                while not done:
                    # Get verbal prompts for this step
                    task_instruction = agent_copy.get_task_instruction()
                    action_history = agent_copy.get_action_history()
                    decision_query = agent_copy.get_decision_query()

                    action_verbal = agent_copy.act()

                    verbal_prompts.append({
                        'task_instruction': task_instruction,
                        'action_history': action_history,
                        'decision_query': decision_query,
                        'label': action_verbal
                    })
                    _, reward, done, info = env_copy.step(action_verbal)
                    if hasattr(agent_copy, 'ag'):
                        action_info = agent_copy.ag.get_actions_guide_info()
                        ag_info.append(action_info)

                    trajectory.append(info['interaction'])
                    action = info['interaction'].mapped_action

                    agent_copy.update(action, reward, info)
                    step_pbar.update(1)

            step_pbar.close()
            trial_pbar.update(1)
            return Trajectory(trajectory), ag_info if hasattr(agent_copy, 'ag') else None, verbal_prompts

        # Update the main progress bar
        if num_threads > 1 and n_trajs > 1:
            with tqdm(total=n_trajs, desc="Collecting trajectories", position=0) as trial_pbar:
                with concurrent.futures.ThreadPoolExecutor(max_workers=num_threads) as executor:
                    futures = [executor.submit(collect_single_trajectory, i, traj_seeds[i], trial_pbar)
                              for i in range(n_trajs)]

                    for future in concurrent.futures.as_completed(futures):
                        trajectory, ag_info, verbal_prompts = future.result()
                        buffer.add(trajectory, ag_info, verbal_prompts)
        else:
            with tqdm(total=n_trajs, desc="Collecting trajectories", position=0) as trial_pbar:
                for i in range(n_trajs):
                    trajectory, ag_info, verbal_prompts = collect_single_trajectory(i, traj_seeds[i], trial_pbar)
                    buffer.add(trajectory, ag_info, verbal_prompts)

        return buffer



import tempfile
import os
import numpy as np
import pytest

from banditbench.sampling.sampler import DatasetBuffer, Trajectory
from banditbench.tasks.mab import BernoulliBandit, VerbalMultiArmedBandit

from banditbench.agents.classics import UCBAgent, LinUCBAgent
from banditbench.agents.llm import (LLMMABAgentRH, LLMMABAgentSH, LLMCBAgentRH, LLMCBAgentRHWithAG, LLMMABAgentSHWithAG,
                                    OracleLLMCBAgentRH, OracleLLMCBAgentRHWithAG, OracleLLMMABAgentSH,
                                    OracleLLMMAbAgentRH, OracleLLMMABAgentSHWithAG)

from banditbench.training.fewshot import FewShot
from banditbench.agents.guides import UCBGuide, LinUCBGuide

from banditbench.tasks.mab.env import Interaction as MABInteraction, VerbalInteraction as MABVerbalInteraction
from banditbench.tasks.cb.env import Interaction as CBInteraction, VerbalInteraction as CBVerbalInteraction

from banditbench.tasks.cb.movielens import MovieLens, MovieLensVerbal


@pytest.fixture
def temp_files():
    files = []
    yield files
    # Cleanup after tests
    for file in files:
        if os.path.exists(file):
            os.remove(file)


def mab_fake_steps(agent, verbal_bandit):
    verbal_bandit.reset()

    _, reward, done, info = verbal_bandit.step('A')

    action = info['interaction'].mapped_action
    agent.update(action, reward, info)

    _, reward, done, info = verbal_bandit.step('B')
    action = info['interaction'].mapped_action
    agent.update(action, reward, info)

    _, reward, done, info = verbal_bandit.step('A')

    action = info['interaction'].mapped_action
    agent.update(action, reward, info)

    _, reward, done, info = verbal_bandit.step('B')
    action = info['interaction'].mapped_action
    agent.update(action, reward, info)


def test_mab_fewshot(temp_files):
    core_bandit = BernoulliBandit(2, 10, [0.2, 0.5], 123)
    verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching")

    # Test LLMMABAgentRH
    agent = LLMMABAgentRH(verbal_bandit, "gpt-3.5-turbo", history_context_len=1000)
    agent.generate = lambda x: "Fake Action"

    buffer = agent.in_context_learn(verbal_bandit, n_trajs=2)
    assert len(buffer) == 2
    assert len(buffer[0].verbal_prompts) == 10

    # Test save/load
    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.json')
    temp_files.append(temp_file.name)
    temp_file.close()

    buffer.dump(temp_file.name)

    # new agent
    agent = LLMMABAgentSH(verbal_bandit, "gpt-3.5-turbo", history_context_len=1000)

    fewshot_agent = FewShot.empower(agent, temp_file.name, 2)
    fewshot_agent.reset()

    mab_fake_steps(fewshot_agent, verbal_bandit)

    fewshot_agent.generate = lambda x: x

    constructed_llm_message = fewshot_agent.act()
    # assert "A video" in constructed_llm_message
    print(constructed_llm_message)


def test_mab_ag_fewshot(temp_files):
    core_bandit = BernoulliBandit(2, 10, [0.2, 0.5], 123)
    verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching")

    # Test LLMMABAgentRH
    ucb_guide = UCBGuide(UCBAgent(core_bandit))
    agent = LLMMABAgentSHWithAG(verbal_bandit, ucb_guide, "gpt-3.5-turbo", history_context_len=1000)
    agent.generate = lambda x: "Fake Action"

    buffer = agent.in_context_learn(verbal_bandit, n_trajs=2)
    assert len(buffer) == 2
    assert len(buffer[0].verbal_prompts) == 10

    # Test save/load
    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.json')
    temp_files.append(temp_file.name)
    temp_file.close()

    buffer.dump(temp_file.name)
    # loaded_buffer = DatasetBuffer.load(temp_file.name)

    # new agent
    agent = LLMMABAgentSH(verbal_bandit, "gpt-3.5-turbo", history_context_len=1000)

    fewshot_agent = FewShot.empower(agent, temp_file.name, 2)
    fewshot_agent.reset()

    mab_fake_steps(fewshot_agent, verbal_bandit)

    fewshot_agent.generate = lambda x: x

    constructed_llm_message = fewshot_agent.act()
    # assert "A video" in constructed_llm_message
    print(constructed_llm_message)


env = None
verbal_env = None


def init_cb_env():
    global env
    global verbal_env

    if env is None:
        env = MovieLens('100k-ratings', num_arms=5, horizon=10, rank_k=5, mode='train',
                        save_data_dir='./tensorflow_datasets/')
    if verbal_env is None:
        verbal_env = MovieLensVerbal(env)

def cb_fake_steps(agent,verbal_env):
    state, info = verbal_env.reset()

    new_state, reward, _, info = verbal_env.step(state, 'Star Wars (1977)')

    action = info['interaction'].mapped_action
    agent.update(state, action, reward, info)

    state = new_state

    new_state, reward, _, info = verbal_env.step(state, 'Fargo (1996)')
    action = info['interaction'].mapped_action
    agent.update(state, action, reward, info)

    state = new_state
    return state

def test_cb_fewshot(temp_files):

    # Test LLMMABAgentRH
    init_cb_env()
    agent = LLMCBAgentRH(verbal_env, "gpt-3.5-turbo", history_context_len=1000)
    agent.generate = lambda x: "Fake Action"

    buffer = agent.in_context_learn(verbal_env, n_trajs=2)
    assert len(buffer) == 2
    assert len(buffer[0].verbal_prompts) == 10

    # Test save/load
    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.json')
    temp_files.append(temp_file.name)
    temp_file.close()

    buffer.dump(temp_file.name)

    # new agent
    agent = LLMCBAgentRH(verbal_env, "gpt-3.5-turbo", history_context_len=1000)

    fewshot_agent = FewShot.empower(agent, temp_file.name, 2)
    fewshot_agent.reset()
    state, _ = verbal_env.reset()

    cb_fake_steps(fewshot_agent, verbal_env)

    fewshot_agent.generate = lambda x: x

    constructed_llm_message = fewshot_agent.act(state)
    # assert "A video" in constructed_llm_message
    print(constructed_llm_message)

def test_cb_ag_fewshot(temp_files):

    # Test LLMMABAgentRH
    init_cb_env()
    guide = LinUCBGuide(LinUCBAgent(env))
    agent = LLMCBAgentRHWithAG(verbal_env, guide, "gpt-3.5-turbo", history_context_len=1000)
    agent.generate = lambda x: "Fake Action"

    buffer = agent.in_context_learn(verbal_env, n_trajs=2)
    assert len(buffer) == 2
    assert len(buffer[0].verbal_prompts) == 10

    # Test save/load
    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.json')
    temp_files.append(temp_file.name)
    temp_file.close()

    buffer.dump(temp_file.name)

    # new agent
    agent = LLMCBAgentRH(verbal_env, "gpt-3.5-turbo", history_context_len=1000)

    fewshot_agent = FewShot.empower(agent, temp_file.name, 2)
    fewshot_agent.reset()
    state, _ = verbal_env.reset()

    cb_fake_steps(fewshot_agent, verbal_env)

    fewshot_agent.generate = lambda x: x

    constructed_llm_message = fewshot_agent.act(state)
    # assert "A video" in constructed_llm_message
    print(constructed_llm_message)



import re
import numpy as np

import os
import csv
import requests
from pathlib import Path

def get_geo_data():
    # First check if file exists in current directory's data folder
    current_dir = os.path.dirname(os.path.abspath(__file__))
    local_data_path = os.path.join(current_dir, 'data', 'geo-data.csv')
    
    if os.path.exists(local_data_path):
        return local_data_path
        
    # If not found locally, use home directory
    home_dir = str(Path.home())
    data_dir = os.path.join(home_dir, '.banditbench', 'data')
    os.makedirs(data_dir, exist_ok=True)

    geo_data_path = os.path.join(data_dir, 'geo-data.csv')

    # Download file if it doesn't exist
    if not os.path.exists(geo_data_path):
        url = 'https://raw.githubusercontent.com/scpike/us-state-county-zip/refs/heads/master/geo-data.csv'
        response = requests.get(url)
        with open(geo_data_path, 'wb') as f:
            f.write(response.content)

    return geo_data_path

movie_genre_to_text = [
    'Action',
    'Adventure',
    'Animation',
    'Children',
    'Comedy',
    'Crime',
    'Documentary',
    'Drama',
    'Fantasy',
    'Film-Noir',
    'Horror',
    'IMAX',
    'Musical',
    'Mystery',
    'Romance',
    'Sci-Fi',
    'Thriller',
    'Unknown',
    'War',
    'Western',
    '(no genres listed)',
]


def safe_decode(value, default='Unknown'):
    if isinstance(value, bytes):
        try:
            return value.decode('utf-8')
        except UnicodeDecodeError:
            try:
                return value.decode('latin-1')
            except UnicodeDecodeError:
                return default
    elif isinstance(value, str):
        try:
            return eval(value).decode('utf-8')
        except:
            return value
    return str(value)


def parse_int_list(s):
    # Remove square brackets and split by whitespace
    numbers = re.findall(r'-?\d+', s)

    # Convert strings to floats
    return [int(num) for num in numbers]


def load_data_files(bandit_type, save_data_dir=None):
    import tensorflow_datasets as tfds

    # Try loading cached files first
    print("loading from tfds")
    if bandit_type == '100k-ratings':
        ratings = tfds.load('movielens/100k-ratings', data_dir=save_data_dir)
    elif bandit_type == '1m-ratings':
        ratings = tfds.load('movielens/1m-ratings', data_dir=save_data_dir)
    else:
        ratings = tfds.load(bandit_type, data_dir=save_data_dir)
        raise ValueError(f"Other movielens datasets not verified yet: {bandit_type}")

    ratings_df = tfds.as_dataframe(ratings['train'])  # there is only one split
    print("data converted to pandas dataframe")

    return ratings_df


def load_movielens_data(ratings_df, top_k_movies=20, seed=1234):
    """Need to return three things:

    1.User index to preference mapping (need to start with an index)
    2. User index to user features
    3. Movie index to movie features

    The missing rating is filled in with 0.
    """
    unique_users = ratings_df['user_id'].unique()

    top_movies = ratings_df['movie_id'].value_counts()[:top_k_movies]
    # unique_movies = ratings_df['movie_id'].unique()
    unique_movies = top_movies.index.values

    # users are shuffled each time
    np.random.seed(seed)
    np.random.shuffle(unique_users)

    user_id_map = {id: i for i, id in enumerate(unique_users)}
    movie_id_map = {id: i for i, id in enumerate(unique_movies)}
    data_matrix = np.zeros((len(unique_users), len(unique_movies)))

    user_idx_to_feats = {}
    movie_idx_to_feats = {}

    for _, row in ratings_df.iterrows():
        i = user_id_map[row['user_id']]
        if row['movie_id'] in movie_id_map:
            j = movie_id_map[row['movie_id']]
            data_matrix[i, j] = row['user_rating']

            if j not in movie_idx_to_feats:
                movie_idx_to_feats[j] = [
                    row['movie_title'],
                    row['movie_genres'],
                ]

        if i not in user_idx_to_feats:
            if 'raw_user_age' in row:
                raw_user_age = row['raw_user_age']
            else:
                raw_user_age = -1

            user_idx_to_feats[i] = [
                row['bucketized_user_age'],
                raw_user_age,
                row['user_gender'],
                row['user_occupation_label'],
                row['user_occupation_text'],
                str(row['user_zip_code']),
            ]

    return data_matrix, user_idx_to_feats, movie_idx_to_feats



import os
import glob
import subprocess


RED = "\033[31m"
GREEN = "\033[32m"
RESET = "\033[0m"


def print_colored(text, color):
    print(f"{color}{text}{RESET}")


# Get the directory where the current module is located
current_dir = os.path.dirname(os.path.abspath(__file__))

# Get a list of all .py files in that directory
py_files = glob.glob(os.path.join(current_dir, "*.py"))

at_least_one_error = False
# Print the list of .py files
for file in py_files:
    filename = os.path.basename(file)
    if filename.startswith("test_"):
        print_colored(f"Running {filename} ...", GREEN)

        result = subprocess.run(["python", file], capture_output=True, text=True)
        if result.returncode == 0:
            print_colored("PASSED", GREEN)
            # print(result.stdout)
        else:
            print_colored("FAILED", RED)
            print(result.stderr)
            at_least_one_error = True

if at_least_one_error:
    exit(1)



import tempfile
import os
import numpy as np
import pytest

from banditbench.sampling.sampler import DatasetBuffer, Trajectory
from banditbench.tasks.mab import BernoulliBandit, VerbalMultiArmedBandit

from banditbench.agents.classics import UCBAgent, LinUCBAgent
from banditbench.agents.guides import UCBGuide, LinUCBGuide
from banditbench.agents.llm import (LLMMABAgentRH, LLMMABAgentSH, LLMCBAgentRH, LLMCBAgentRHWithAG, LLMMABAgentSHWithAG,
                                    OracleLLMCBAgentRH, OracleLLMCBAgentRHWithAG, OracleLLMMABAgentSH,
                                    OracleLLMMAbAgentRH, OracleLLMMABAgentSHWithAG)


from banditbench.tasks.mab.env import Interaction as MABInteraction, VerbalInteraction as MABVerbalInteraction
from banditbench.tasks.cb.env import Interaction as CBInteraction, VerbalInteraction as CBVerbalInteraction

from banditbench.tasks.cb.movielens import MovieLens, MovieLensVerbal


@pytest.fixture
def temp_files():
    files = []
    yield files
    # Cleanup after tests
    for file in files:
        if os.path.exists(file):
            os.remove(file)


def test_save_and_load(temp_files):
    core_bandit = BernoulliBandit(2, 10, [0.5, 0.2], 123)
    verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching")

    # Create a temporary file
    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.json')
    temp_files.append(temp_file.name)
    temp_file.close()

    # Create test data
    buffer = DatasetBuffer()

    _, reward, done, info = verbal_bandit.step('A')
    inter1 = info['interaction']

    _, reward, done, info = verbal_bandit.step('B')
    inter2 = info['interaction']

    test_traj = inter1 + inter2

    buffer.append(test_traj)

    # Test save
    buffer.dump(temp_file.name)
    assert os.path.exists(temp_file.name)

    # Test load
    loaded_buffer = DatasetBuffer.load(temp_file.name)
    assert len(loaded_buffer) == 1

    loaded_traj = loaded_buffer[0]
    assert len(loaded_traj) == 2
    assert loaded_traj[0].raw_action == inter1.raw_action
    assert loaded_traj[0].reward == inter1.reward


def test_multiple_trajectories(temp_files):
    core_bandit = BernoulliBandit(2, 10, [0.5, 0.2], 123)
    verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching")

    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.json')
    temp_files.append(temp_file.name)
    temp_file.close()

    buffer = DatasetBuffer()
    n_trajectories = 5

    # Add multiple trajectories
    for _ in range(n_trajectories):
        _, reward, done, info = verbal_bandit.step('A')
        inter1 = info['interaction']
        _, reward, done, info = verbal_bandit.step('B')
        inter2 = info['interaction']

        traj = inter1 + inter2
        buffer.append(traj)

    # Save and load
    buffer.save(temp_file.name)
    loaded_buffer = DatasetBuffer.load(temp_file.name)

    # Verify
    assert len(loaded_buffer) == n_trajectories
    for traj in loaded_buffer:
        assert len(traj) == 2
        assert isinstance(traj[0], MABVerbalInteraction)
        assert isinstance(traj[1], MABVerbalInteraction)


def test_databuffer_slicing():
    core_bandit = BernoulliBandit(2, 10, [0.5, 0.2], 123)
    verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching")

    buffer = DatasetBuffer()
    n_trajectories = 3

    # Add multiple trajectories
    for _ in range(n_trajectories):
        _, reward, done, info = verbal_bandit.step('A')
        inter1 = info['interaction']
        _, reward, done, info = verbal_bandit.step('B')
        inter2 = info['interaction']
        _, reward, done, info = verbal_bandit.step('B')
        inter3 = info['interaction']

        traj = inter1 + inter2 + inter3
        buffer.append(traj)

    # Test slicing
    sliced_buffer = buffer[:1]
    assert len(sliced_buffer) == 1
    assert sliced_buffer[0] == buffer[0]
    sliced_buffer = buffer[1:2]
    assert sliced_buffer[0] == buffer[1]
    assert len(sliced_buffer) == 1
    sliced_buffer = buffer[0:2]
    assert len(sliced_buffer) == 2
    assert sliced_buffer[0] == buffer[0]
    assert sliced_buffer[1] == buffer[2]


def test_mab_sampling(temp_files):
    core_bandit = BernoulliBandit(2, 200, [0.2, 0.5], 123)

    agent = UCBAgent(core_bandit)
    dataset = agent.in_context_learn(core_bandit, 100)
    assert len(dataset) == 100
    dataset.plot_performance()

    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.json')
    temp_files.append(temp_file.name)
    temp_file.close()

    dataset.save(temp_file.name)
    loaded_buffer = DatasetBuffer.load(temp_file.name)

    assert len(loaded_buffer) == 100


def test_mab_ag_sampling(temp_files):
    core_bandit = BernoulliBandit(2, 200, [0.2, 0.5], 123)

    guide = UCBGuide(UCBAgent(core_bandit))
    dataset = guide.in_context_learn(core_bandit, 20)
    assert len(dataset) == 20

    assert len(dataset.action_infos) == 20

    # the 1st trajectory, 1st interaction, 1st action's info
    print(dataset.action_infos[0][0][0].to_str())

    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.json')
    temp_files.append(temp_file.name)
    temp_file.close()

    dataset.save(temp_file.name)
    loaded_buffer = DatasetBuffer.load(temp_file.name)

    assert len(loaded_buffer) == 20


env = None
verbal_env = None


def init_cb_env():
    global env
    global verbal_env

    if env is None:
        env = MovieLens('100k-ratings', num_arms=5, horizon=10, rank_k=5, mode='train',
                        save_data_dir='./tensorflow_datasets/')
    if verbal_env is None:
        verbal_env = MovieLensVerbal(env)


def test_mab_llm_rh_sampling(temp_files):
    # Setup environment
    core_bandit = BernoulliBandit(2, 10, [0.2, 0.5], 123)
    verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching")

    # Test LLMMABAgentRH
    agent = LLMMABAgentRH(verbal_bandit, "gpt-3.5-turbo", history_context_len=1000)
    buffer = agent.in_context_learn(verbal_bandit, n_trajs=2)
    assert len(buffer) == 2
    assert len(buffer[0].verbal_prompts) == 10

    # Test save/load
    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.json')
    temp_files.append(temp_file.name)
    temp_file.close()

    buffer.dump(temp_file.name)
    loaded_buffer = DatasetBuffer.load(temp_file.name)
    assert len(loaded_buffer) == 2


def test_mab_llm_sh_sampling(temp_files):
    # Setup environment
    core_bandit = BernoulliBandit(2, 10, [0.2, 0.5], 123)
    verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching")

    # Test LLMMABAgentSH
    agent = LLMMABAgentSH(verbal_bandit, "gpt-3.5-turbo", history_context_len=1000)
    agent.generate = lambda x: ""  # so that the LLM is not triggered
    buffer = agent.in_context_learn(verbal_bandit, n_trajs=2)
    assert len(buffer) == 2
    assert len(buffer[0].verbal_prompts) == 10

    # Test save/load
    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.json')
    temp_files.append(temp_file.name)
    temp_file.close()

    buffer.dump(temp_file.name)
    loaded_buffer = DatasetBuffer.load(temp_file.name)
    assert len(loaded_buffer) == 2


def test_cb_llm_rh_sampling(temp_files):
    # Test LLMCBAgentRH
    init_cb_env()
    agent = LLMCBAgentRH(verbal_env, "gpt-3.5-turbo", history_context_len=1000)
    agent.generate = lambda x: ""  # so that the LLM is not triggered
    buffer = agent.in_context_learn(verbal_env, n_trajs=2)
    assert len(buffer) == 2
    assert len(buffer[0].verbal_prompts) == 10

    # Test save/load
    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.json')
    temp_files.append(temp_file.name)
    temp_file.close()

    buffer.dump(temp_file.name)
    loaded_buffer = DatasetBuffer.load(temp_file.name)
    assert len(loaded_buffer) == 2


def test_cb_llm_rh_with_ag_sampling(temp_files):
    # Test LLMCBAgentRHWithAG
    init_cb_env()
    ucb_guide = LinUCBGuide(LinUCBAgent(env))
    agent = LLMCBAgentRHWithAG(verbal_env, ucb_guide, "gpt-3.5-turbo", history_context_len=1000)
    agent.generate = lambda x: ""  # so that the LLM is not triggered
    buffer = agent.in_context_learn(verbal_env, n_trajs=2)
    assert len(buffer) == 2
    assert len(buffer[0].verbal_prompts) == 10
    assert buffer[0].action_info is not None

    # Test save/load
    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.json')
    temp_files.append(temp_file.name)
    temp_file.close()

    buffer.dump(temp_file.name)
    loaded_buffer = DatasetBuffer.load(temp_file.name)
    assert len(loaded_buffer) == 2


def test_mab_llm_sh_with_ag_sampling(temp_files):
    # Setup environment
    core_bandit = BernoulliBandit(2, 10, [0.2, 0.5], 123)
    verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching")

    # Test LLMMABAgentSHWithAG
    ucb_guide = UCBGuide(UCBAgent(core_bandit))
    agent = LLMMABAgentSHWithAG(verbal_bandit, ucb_guide, "gpt-3.5-turbo", history_context_len=1000)
    agent.generate = lambda x: ""  # so that the LLM is not triggered
    buffer = agent.in_context_learn(verbal_bandit, n_trajs=2)
    assert len(buffer) == 2
    assert len(buffer[0].verbal_prompts) == 10
    assert buffer[0].action_info is not None

    # Test save/load
    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.json')
    temp_files.append(temp_file.name)
    temp_file.close()

    buffer.dump(temp_file.name)
    loaded_buffer = DatasetBuffer.load(temp_file.name)
    assert len(loaded_buffer) == 2


def test_mab_llm_sh_with_ag_sampling(temp_files):
    # Setup environment
    core_bandit = BernoulliBandit(2, 10, [0.2, 0.5], 123)
    verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching")

    # Test LLMMABAgentSHWithAG
    ucb_guide = UCBGuide(UCBAgent(core_bandit))
    oracle = UCBAgent(core_bandit)
    agent = OracleLLMMABAgentSHWithAG(verbal_bandit, ucb_guide, oracle)

    agent.generate = lambda x: ""  # so that the LLM is not triggered
    buffer = agent.in_context_learn(verbal_bandit, n_trajs=2)
    assert len(buffer) == 2
    assert len(buffer[0].verbal_prompts) == 10
    assert buffer[0].action_info is not None

    # Test save/load
    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.json')
    temp_files.append(temp_file.name)
    temp_file.close()

    buffer.to_sft_data(temp_file.name)
    data = buffer.to_sft_data()
    print(data[0][4])



import os
from typing import List, Optional

from banditbench.tasks.types import State
from banditbench.utils import dedent
from banditbench.tasks.scenario import CBScenario

class ScenarioUtil:
    """
    Inheriting from this class unifies the subclass' default __init__ method
    """

    def __init__(self, action_names, num_actions: int,
                 seed: Optional[int] = None):
        super().__init__(num_actions=num_actions, action_names=action_names, action_unit=self.action_unit,
                         base_description=self.base_description, detailed_description=self.detailed_description,
                         query_prompt=self.query_prompt, seed=seed)


class MovieLensScenario(ScenarioUtil, CBScenario):
    name = 'movie'
    action_unit = 'movie'
    reward_unit = 'User returned a rating of'
    max_reward = '5'

    base_description = dedent(
        """You are an AI movie recommendation assistant for a streaming platform powered by a bandit algorithm that offers a wide variety of films from different studios and genres.
          There are {} unique movies you can recommend, named {}
          When a user visits the streaming platform, you assess their demographic description to choose a movie to suggest.
          You aim to match the user with movies they are most likely to watch and enjoy.
          Each time a user watches a recommended movie, you adjust your recommendation algorithms to better predict and meet future user preferences.
          Your goal is to enhance the user's viewing experience by providing personalized and engaging movie suggestions.
        """
    )

    detailed_description = dedent(
        """You are an AI movie recommendation assistant for a streaming platform powered by a bandit algorithm that offers a wide variety of films from different studios and genres.
          There are {} unique movies you can recommend, named {}
          When a user visits the streaming platform, you assess their demographic description to choose a movie to suggest.
          You aim to match the user with movies they are most likely to watch and enjoy.
          Each time a user watches a recommended movie, you adjust your recommendation algorithms to better predict and meet future user preferences.
          Your goal is to enhance the user's viewing experience by providing personalized and engaging movie suggestions.
          
          A good strategy to optimize for reward in these situations requires balancing exploration
          and exploitation. You need to explore to try out different movies and find those
          with high rewards, but you also have to exploit the information that you have to
          accumulate rewards.
        """
    )

    query_prompt = ("\n\nYou have a new user: PLEASE"
                    " RESPOND ONLY WITH A CHOICE of MOVIES LISTED ABOVE AND NO TEXT"
                    " EXPLANATION.\n\nContext: {feature}\n")

    fewshot_prompt = ('Here are some examples of optimal actions for different users.'
                      ' Use them as hints to help you come up with better actions.\n')

    action_names: List[str]

    def get_instruction(self, version="base"):
        """We add the few-shot examples in here"""
        if version == 'base':
            prompt = self.base_description.format(
                len(self.action_names), '\n' + ',\n'.join(self.action_names) + '\n'
            )
        elif version == 'detailed':
            prompt = self.detailed_description.format(
                len(self.action_names), '\n' + ',\n'.join(self.action_names) + '\n'
            )
        else:
            raise ValueError(f'Unknown description version {version}')

        return prompt

    def get_query_prompt(self, state: State, side_info: Optional[str] = None):
        assert type(
            state.feature_text) is str, "Verbal bandit produces string as feature, make sure you are using the right bandit class"
        prompt = self.query_prompt.format(feature=state.feature_text)
        if side_info is not None:
            prompt += f"Side Information for decision making:\n{side_info}"

        prompt += "Action: "

        return prompt



from typing import List, Dict, Any, Union, Tuple
import scipy
import numpy as np
from pydantic import BaseModel
from banditbench.agents.classics import MABAgent, CBAgent, UCBAgent, ThompsonSamplingAgent, GreedyAgent, LinUCBAgent
from banditbench.tasks.cb.env import State
from banditbench.agents.types import ActionInfo, ActionInfoField
from banditbench.sampling.sampler import VerbalGuideSampleBase


class VerbalGuideBase(VerbalGuideSampleBase):
    # VerbalGuide can be retrieved in two ways:
    # it's a verbal analog of a Q-function
    # Q(s=None, a) # MAB
    # Q(s, a) # CB

    def __init__(self, agent: Union[MABAgent, CBAgent]):
        self.agent = agent

    def parse_args_is_agent(self, *args: List[str]) -> Tuple[bool, Any]:
        for arg in args:
            if hasattr(arg, 'act') and hasattr(arg, 'update'):
                return True, arg
            elif hasattr(arg, 'horizon'):
                return False, arg

        raise ValueError("Unknown argument type. Please pass in the environment or the agent.")

    def get_action_guide_info(self, arm: int) -> ActionInfo:
        raise NotImplementedError("Only for MAB agents")

    def get_actions_guide_info(self) -> List[ActionInfo]:
        raise NotImplementedError("Only for MAB agents")

    def get_state_action_guide_info(self, state: State, arm: int) -> ActionInfo:
        raise NotImplementedError("Only for RL and CB agents")

    def get_state_actions_guide_info(self, state: State) -> List[ActionInfo]:
        raise NotImplementedError("Only for RL and CB agents")


class UCBGuide(VerbalGuideBase):
    # takes in UCBAgent and then return info on each arm (a block of text)
    def __init__(self, *args, **kwargs):
        """
        :param args: Should be either env or agent. Can pass in flags to agent through kwargs (like "alpha=0.5")
        :param agent:
        """
        is_agent, arg = self.parse_args_is_agent(*args)
        if is_agent:
            super().__init__(arg)
        else:
            super().__init__(UCBAgent(arg, **kwargs))

    def get_actions_guide_info(self) -> List[ActionInfo]:
        actions_info = []
        for arm in range(self.agent.k_arms):
            arm_info = self.get_action_guide_info(arm)
            actions_info.append(arm_info)

        assert len(actions_info) == self.agent.k_arms
        return actions_info

    def get_action_guide_info(self, arm: int) -> ActionInfo:
        exploration_bonus = self.agent.calculate_exp_bonus(arm) if self.agent.arms[arm] > 0 else "inf"
        exp_bonus_guide = ActionInfoField(info_name='exploration bonus', value=exploration_bonus)

        exploitation_value = self.agent.calculate_exp_value(arm) if self.agent.arms[arm] > 0 else 0
        exp_value_guide = ActionInfoField(info_name='exploitation value', value=exploitation_value)
        return exp_bonus_guide + exp_value_guide


class GreedyGuide(VerbalGuideBase):
    def __init__(self, *args, **kwargs):
        """
        :param args: Should be either env or agent. Can pass in flags to agent through kwargs (like "alpha=0.5")
        :param agent:
        """
        is_agent, arg = self.parse_args_is_agent(*args)
        if is_agent:
            super().__init__(arg)
        else:
            super().__init__(GreedyAgent(arg, **kwargs))

    def get_actions_guide_info(self) -> List[ActionInfo]:
        actions_info = []
        for arm in range(self.agent.k_arms):
            arm_info = self.get_action_guide_info(arm)
            actions_info.append(arm_info)
        return actions_info

    def get_action_guide_info(self, arm: int) -> ActionInfo:
        exploitation_value = self.agent.calculate_exp_value(arm) if self.agent.arms[arm] > 0 else 0
        exp_value_guide = ActionInfoField(info_name='exploitation value', value=exploitation_value)
        arm_info = ActionInfo(action_info_fields=[exp_value_guide])
        return arm_info


class ThompsonSamplingGuide(VerbalGuideBase):
    def __init__(self, *args, **kwargs):
        """
        :param args: Should be either env or agent. Can pass in flags to agent through kwargs (like "alpha=0.5")
        :param agent:
        """
        is_agent, arg = self.parse_args_is_agent(*args)
        if is_agent:
            super().__init__(arg)
        else:
            super().__init__(ThompsonSamplingAgent(arg, **kwargs))

    def get_actions_guide_info(self) -> List[ActionInfo]:
        actions_info = []
        for arm in range(self.agent.k_arms):
            arm_info = self.get_action_guide_info(arm)
            actions_info.append(arm_info)

        assert len(actions_info) == len(self.agent.actions)
        return actions_info

    def get_action_guide_info(self, arm: int) -> ActionInfo:
        alpha = self.agent.alpha[arm]
        beta = self.agent.beta[arm]
        p = scipy.stats.beta.rvs(self.agent.alpha[arm], self.agent.beta[arm])
        alpha_guide = ActionInfoField(info_name='alpha', value=alpha,
                                      info_template='prior beta distribution(alpha={:.2f}')
        beta_guide = ActionInfoField(info_name='beta', value=beta, info_template='beta={:.2f})')
        probability_guide = ActionInfoField(info_name='probability', value=p,
                                            info_template='posterior bernoulli p={:.2f}')
        return alpha_guide + beta_guide + probability_guide


class LinUCBGuide(VerbalGuideBase):
    def __init__(self, *args, **kwargs):
        """
        :param args: Should be either env or agent. Can pass in flags to agent through kwargs (like "alpha=0.5")
        :param agent:
        """
        is_agent, arg = self.parse_args_is_agent(*args)
        if is_agent:
            super().__init__(arg)
        else:
            super().__init__(LinUCBAgent(arg, **kwargs))

    def get_state_action_guide_info(self, state: State, arm: int) -> ActionInfo:
        a = arm
        context = state.feature

        A_inv = np.linalg.inv(self.agent.A[a])
        theta = A_inv.dot(self.agent.b[a])
        exploration_bonus = self.agent.alpha * np.sqrt(context.T.dot(A_inv).dot(context))
        # exploitation_value = theta.T.dot(context)[0, 0]
        exploitation_value = theta.T.dot(context)[0]

        exp_bonus_guide = ActionInfoField(info_name='exploration bonus', value=exploration_bonus)
        exp_value_guide = ActionInfoField(info_name='exploitation value', value=exploitation_value)

        return exp_bonus_guide + exp_value_guide

    def get_state_actions_guide_info(self, state: State) -> List[ActionInfo]:
        actions_info = []
        for arm in range(self.agent.k_arms):
            arm_info = self.get_state_action_guide_info(state, arm)
            actions_info.append(arm_info)

        return actions_info


class VerbalGuideBuilder:
    pass


class VerbalGuide:
    pass



from typing import Optional, List, Any, Dict, Union

from banditbench.agents.guides import VerbalGuideBase, UCBGuide, LinUCBGuide, ActionInfo
from banditbench.tasks.types import State, Info

import banditbench.tasks.cb as cb
import banditbench.tasks.mab as mab

from banditbench.agents.types import MABAgent, CBAgent, ActionInfo


class ModelContextLayer:
    interaction_history: List[Union[mab.VerbalInteraction, cb.VerbalInteraction]]
    history_context_len: int

    def __init__(self, history_context_len: int):
        self.interaction_history = []
        self.history_context_len = history_context_len

    def represent_interaction_context(self, action_names: List[str], action_unit: str,
                                      history_len: int) -> str:
        """Get formatted history for LLM prompt."""
        # Implement history formatting
        raise NotImplementedError

    def reset(self):
        self.interaction_history = []


class RawContextLayerMAB(ModelContextLayer):
    """Formats raw interaction history for LLM prompt."""

    def represent_interaction_context(self, action_names: List[str], action_unit: str,
                                      history_len: int) -> str:
        if len(self.interaction_history) == 0:
            return ""

        # remember to handle state
        history_len = min(history_len, len(self.interaction_history))
        snippet = ""
        for exp in self.interaction_history[-history_len:]:
            snippet += f"\n{exp.feedback}"  # MAB feedback contains {action_name} {reward} already

        return snippet


class RawContextLayerCB(ModelContextLayer):
    def represent_interaction_context(self, action_names: List[str], action_unit: str,
                                      history_len: int) -> str:
        if len(self.interaction_history) == 0:
            return ""

        # remember to handle state
        history_len = min(history_len, len(self.interaction_history))
        snippet = ""
        for exp in self.interaction_history[-history_len:]:
            snippet += f"\nContext: {exp.state.feature_text}"
            snippet += f"\nAction: {exp.mapped_action_name}"  # this is to replicate the same style as the paper
            snippet += f"\nReward: {exp.reward}\n"

        return snippet


class SummaryContextLayerMAB(ModelContextLayer):
    """Summarizes interaction history for LLM prompt."""

    def represent_interaction_context(self, action_names: List[str], action_unit: str,
                                      history_len: int) -> str:
        """
        Note that this function can work with either MAB or CB
        But for CB, it is not summarizing on the state level
        """
        # we traverse through the whole history to summarize
        if len(self.interaction_history) == 0:
            return ""

        # compute basic statistics, for each action name
        # frequency, mean reward

        n_actions = [0] * len(action_names)
        action_rewards = [0] * len(action_names)

        for exp in self.interaction_history:
            idx = action_names.index(exp.mapped_action_name)
            n_actions[idx] += 1
            action_rewards[idx] += exp.reward

        snippet = ""
        for action_name, n, total_r in zip(action_names, n_actions, action_rewards):
            reward = total_r / (n + 1e-6)
            snippet += (
                f"\n{action_name} {action_unit}, {n} times, average"
                f" reward {reward:.2f}"
            )

        return snippet


class SummaryAlgoGuideLayerMAB(SummaryContextLayerMAB):
    """Provides algorithm guidance text for LLM prompt."""
    ag: UCBGuide
    ag_info_history: List[List[ActionInfo]]  # storing side information

    def __init__(self, ag: UCBGuide, history_context_len: int):
        super().__init__(history_context_len)
        self.ag = ag
        self.ag_info_history = []
        assert type(ag) is UCBGuide, "Only UCBGuide works with SummaryHistory -- since the summary is per action level."

    def update_algorithm_guide(self, action: int, reward: float, info: Dict[str, Any]) -> None:
        """Enhance update to include algorithm guide updates."""
        # First call the parent class's update
        self.ag.agent.update(action, reward, info)

    def update_info_history(self, action_info: List[ActionInfo]) -> None:
        self.ag_info_history.append(action_info)

    def reset(self):
        super().reset()  # HistoryFunc.reset()
        self.ag.agent.reset()

    def represent_interaction_context(self, action_names: List[str], action_unit: str,
                                      history_len: int) -> str:
        """
        Note that this function can work with either MAB or CB
        But for CB, it is not summarizing on the state level
        """
        # we traverse through the whole history to summarize
        if len(self.interaction_history) == 0:
            return ""

        n_actions = [0] * len(action_names)
        action_rewards = [0] * len(action_names)

        for exp in self.interaction_history:
            idx = action_names.index(exp.mapped_action_name)
            n_actions[idx] += 1
            action_rewards[idx] += exp.reward

        snippet, action_idx = "", 0
        for action_name, n, total_r in zip(action_names, n_actions, action_rewards):
            reward = total_r / (n + 1e-6)
            snippet += (
                    f"\n{action_name} {action_unit}, {n} times, average"
                    f" reward {reward:.2f}" + " " + self.ag.get_action_guide_info(action_idx).to_str()
            )
            action_idx += 1

        return snippet


class RawContextAlgoGuideLayerCB(RawContextLayerCB):
    """Provides algorithm guidance text for LLM prompt."""
    ag: LinUCBGuide
    ag_info_history: List[List[ActionInfo]]  # storing side information

    def __init__(self, ag: LinUCBGuide, history_context_len: int):
        super().__init__(history_context_len)
        self.ag_info_history = []
        self.ag = ag
        assert type(ag) is LinUCBGuide, "The information is provided per context, per action"

    def reset(self):
        super().reset()  # HistoryFunc.reset()
        self.ag.agent.reset()
        self.ag_info_history = []

    def update_algorithm_guide(self, state: State, action: int, reward: float, info: Dict[str, Any]) -> None:
        """Enhance update to include algorithm guide updates."""
        # First call the parent class's update
        self.ag.agent.update(state, action, reward, info)

    def update_info_history(self, action_info: List[ActionInfo]) -> None:
        self.ag_info_history.append(action_info)

    def represent_interaction_context(self, action_names: List[str], action_unit: str,
                                      history_len: int) -> str:
        if len(self.interaction_history) == 0:
            return ""

        # remember to handle state
        history_len = min(history_len, len(self.interaction_history))
        snippet = ""
        for exp, ag_info in zip(self.interaction_history[-history_len:], self.ag_info_history[-history_len:]):
            snippet += f"\nContext: {exp.state.feature_text}"
            snippet += f"\nSide Information for decision making:"
            for i, action_info in enumerate(ag_info):
                # normal format
                # snippet += '\n' + action_names[i].split(") (")[0] + ")" + ": " + action_info.to_str()

                # JSON-like format used in the paper
                snippet += '\n{\"' + action_names[i] + ": " + action_info.to_str(
                    json_fmt=True) + "}"
            snippet += f"\nAction: {exp.mapped_action_name}"
            snippet += f"\nReward: {exp.reward}\n"

        return snippet



from banditbench.tasks.mab.scenarios import ButtonPushing

def test_button_pushing():
    scenario = ButtonPushing(5)
    print()
    print(scenario.get_instruction("base"))
    print()
    print(scenario.get_instruction("detailed"))

test_button_pushing()


from pydantic import BaseModel
from typing import Dict, Any, Tuple, Union, List, Optional
import numpy as np
from banditbench.tasks.scenario import MABScenario
from banditbench.tasks.mab.scenarios import ButtonPushing, OnlineAds, VideoWatching, ClothesShopping
from banditbench.tasks.types import Action, ExpectedReward, InteractionBase
from banditbench.tasks.env import Bandit, VerbalBandit

BernArmParam = float
GaussianArmParam = Tuple[float, float]
BanditArmParam = Union[BernArmParam, GaussianArmParam]


class Interaction(BaseModel, InteractionBase):
    action: Action
    reward: float  # observed reward
    expected_reward: ExpectedReward  # this information is hidden from the agent
    is_random: Union[bool, None] = None

    def __init__(self, action: Action, reward: float,
                 expected_reward: ExpectedReward, is_random: Union[bool, None] = None) -> None:
        super().__init__(action=action, reward=reward,
                         expected_reward=expected_reward, is_random=is_random)


class VerbalInteraction(BaseModel, InteractionBase):
    raw_action: Action
    mapped_action: Action
    mapped_action_name: Action
    reward: float
    expected_reward: ExpectedReward
    feedback: str
    is_random: Union[bool, None] = None

    def __init__(self, raw_action: Action, mapped_action: Action, mapped_action_name: Action,
                 reward: float, expected_reward: ExpectedReward, feedback: str,
                 is_random: Union[bool, None] = None) -> None:
        super().__init__(raw_action=raw_action, mapped_action=mapped_action,
                         mapped_action_name=mapped_action_name,
                         reward=reward, expected_reward=expected_reward, feedback=feedback,
                         is_random=is_random)


class MultiArmedBandit(Bandit):
    arm_params: List[BanditArmParam]
    history: List[Interaction]
    instance_hardness: float  # this is the Delta_min or "gap" in the paper

    def __init__(self, num_arms: int, horizon: int, arm_params: List[BanditArmParam], seed: Optional[int] = None,
                 instance_hardness: float = 0.0, shuffle_at_reset=False) -> None:
        """
        :param shuffle_at_reset: Setting this will shuffle the underlying arm param distribution at each reset.
        """
        self.num_arms = num_arms
        self.horizon = horizon
        self.arm_params = arm_params
        self.set_seed(seed)
        self.instance_hardness = instance_hardness  # this is the optimality gap -- the lower this number is, the harder an instance is
        self.validate()
        if seed is not None:
            self.shuffle_arms()
        self.initialize_defaults()
        self.shuffle_at_reset = shuffle_at_reset

    def initialize_defaults(self) -> None:
        """Initialize default values for type-annotated attributes, but are instance variables"""
        self.history = []
        self.h = 0

    def reward_fn(self, action: int) -> float:
        """In a stochastic bandit, this samples from R_t for the given action"""
        raise NotImplementedError

    def expected_reward(self, action: int) -> float:
        """In a stochastic bandit, this retrieves the E[R_t] for the given action"""
        raise NotImplementedError

    def shuffle_arms(self) -> None:
        self.np_random.shuffle(self.arm_params)

    def validate(self) -> None:
        assert len(
            self.arm_params) == self.num_arms, f"Number of arm parameters {len(self.arm_params)} does not match number of arms {self.num_arms}."
        assert self.horizon > 0, f"Horizon {self.horizon} must be positive."

    def step(self, action: int) -> Tuple[None, float, bool, Dict[str, Any]]:
        assert action >= 0 and action < self.num_arms, f"Action {action} is not in the action space [{0}, {self.num_arms}]."
        if self.h >= self.horizon:
            raise ValueError(
                "Episode is done. Call env.reset() to start a new episode."
            )

        info = {'arm_param': self.arm_params[action]}

        self.h += 1
        done = self.h == self.horizon
        reward = self.reward_fn(action)
        interaction = Interaction(action, reward, self.expected_reward(action), is_random=False)
        self.history.append(interaction)

        info['interaction'] = interaction
        info['is_random'] = False

        return None, reward, done, info

    def reset(self, seed: Optional[int] = None) -> None:
        self.history = []
        self.h = 0
        if seed is not None:
            self.set_seed(seed)

        if self.shuffle_at_reset:
            self.shuffle_arms()


class BernoulliBandit(MultiArmedBandit):

    def __init__(
            self,
            num_arms: int,
            horizon: int,
            arm_params: List[BernArmParam],
            seed: Optional[int] = None,  # might remove this
    ):
        # instance_hardness is computed by getting the highest arm and subtracting the second highest arm
        instance_hardness = max(arm_params) - sorted(arm_params)[-2]
        super(BernoulliBandit, self).__init__(num_arms=num_arms, horizon=horizon, arm_params=arm_params, seed=seed,
                                              instance_hardness=instance_hardness)

    def reward_fn(self, action: int) -> float:
        return 1 if self.np_random.uniform(0, 1) < self.arm_params[action] else 0

    def expected_reward(self, action: int) -> float:
        return self.arm_params[action]

    @property
    def name(self) -> str:
        # b_vid_arms5_easy
        instance_hardness_str = f"{self.instance_hardness:.1f}".replace(".", "_")
        return f"b_arms{self.num_arms}_difficulty_{instance_hardness_str}"


def compute_simple_kl(p_mu, q_mu, sigma):
    return (sigma ** 2 + (p_mu - q_mu) ** 2) / (2 * sigma ** 2) - 1 / 2


def compute_general_kl(p_mu, p_sigma, q_mu, q_sigma):
    return np.log(q_sigma / p_sigma) + (p_sigma ** 2 + (p_mu - q_mu) ** 2) / (2 * q_sigma ** 2) - 0.5


class GaussianBandit(MultiArmedBandit):

    def __init__(self, num_arms: int, horizon: int, arm_params: List[GaussianArmParam],
                 seed: Optional[int] = None) -> None:
        # to make difficulty easy to assess, we require all arms to have the same variance
        assert all(p[1] == arm_params[0][1] for p in arm_params), "All arms must have the same variance"
        # then we compute the instance hardness as the KL divergence between the best and second best arm
        instance_hardness = compute_simple_kl(max(arm_params)[0], sorted(arm_params)[-2][0], max(arm_params)[1])
        super(GaussianBandit, self).__init__(num_arms=num_arms, horizon=horizon, arm_params=arm_params, seed=seed,
                                             instance_hardness=instance_hardness)

    def reward_fn(self, action: int) -> float:
        return self.np_random.normal(self.arm_params[action][0], self.arm_params[action][1])

    def expected_reward(self, action: int) -> float:
        return self.arm_params[action][0]

    @property
    def name(self) -> str:
        # b_vid_arms5_easy
        return f"g_arms{self.num_arms}_difficulty_{self.instance_hardness:.2f}"


class VerbalMultiArmedBandit(VerbalBandit):
    history: List[VerbalInteraction]

    def __init__(self,
                 core_bandit: MultiArmedBandit,
                 bandit_scenario: Union[str, MABScenario, type],
                 # ===== arguments for bandit_scenario_cls =====
                 scenario_seed: Optional[int] = None,
                 instruction_type: str = "detailed") -> None:
        """
        bandit_scenario: Can be one of three types of inputs: a string, a instantiated BanditScenario class, or the class constructor itself
        """
        self.num_arms = core_bandit.num_arms
        self.horizon = core_bandit.horizon
        self.core_bandit = core_bandit
        self.initialize_defaults()

        self.instruction_type = instruction_type

        if isinstance(bandit_scenario, str):
            assert bandit_scenario in ["ButtonPushing", "OnlineAds", "VideoWatching",
                                       "ClothesShopping"], "Unknown bandit scenario"
            self.bandit_scenario = eval(bandit_scenario)(num_actions=self.num_arms,
                                                         seed=scenario_seed)
        elif isinstance(bandit_scenario, type):
            # noinspection PyCallingNonCallable
            self.bandit_scenario = bandit_scenario(num_actions=self.num_arms,
                                                   seed=scenario_seed)
        elif isinstance(bandit_scenario, MABScenario):
            self.bandit_scenario = bandit_scenario
        else:
            raise ValueError("Unknown bandit scenario")

    def initialize_defaults(self) -> None:
        self.history = []

    def reset(self, seed: Optional[int] = None) -> Any:
        self.core_bandit.reset(seed)

    def get_query_prompt(self, *args, **kwargs) -> str:
        return self.bandit_scenario.get_query_prompt()

    def get_task_instruction(self, *args, **kwargs) -> str:
        verbal_instruction = self.bandit_scenario.get_instruction(self.instruction_type)
        return verbal_instruction

    @property
    def action_names(self) -> List[str]:
        return self.bandit_scenario.action_names

    def step(self, action: str) -> Tuple[None, float, bool, Dict[str, Any]]:
        """
        For Verbal environment, reward (numerical) also gets verbalized into feedback (text)
        """
        assert type(action) == str, "Action must be a string for VerbalBandit"
        # Find matching action name, accounting for case and whitespace
        raw_action = action
        action = action.strip().lower()
        action_names = [name.strip().lower() for name in self.bandit_scenario.action_names]
        try:
            action_index = action_names.index(action)
            is_random = False
        except ValueError:
            is_random = True
            action_index = int(self.core_bandit.np_random.integers(0, len(self.bandit_scenario.action_names)))

        # now get the reward
        state, reward, done, info = self.core_bandit.step(action_index)
        assert state is None, "State should be None for MultiArmedBandit"

        feedback = self.verbalize_feedback(self.action_names[action_index], reward)

        interaction = VerbalInteraction(raw_action, action_index, self.action_names[action_index],
                                        reward, self.core_bandit.expected_reward(action_index),
                                        feedback, is_random)
        self.history.append(interaction)

        info['interaction'] = interaction
        info['is_random'] = is_random

        return None, reward, done, info

    def verbalize_feedback(self, action_name: Action, reward: float) -> str:
        # :.3f
        assert type(action_name) == str, "Action name must be a string"
        feedback = "{} {}, reward {}".format(action_name, self.bandit_scenario.action_unit,
                                             reward)
        return feedback

    @property
    def name(self) -> str:
        # b_vid_arms5_easy
        return self.core_bandit.name + "_" + self.bandit_scenario.name



import numpy as np
from banditbench.tasks.mab.env import Interaction, VerbalInteraction
from banditbench.tasks.mab.env import BernoulliBandit, GaussianBandit, VerbalMultiArmedBandit


def create_small_gap_bernoulli_bandit(num_arms, horizon, seed=None):
    small_gap_delta = 0.2
    means = [0.5 - small_gap_delta / 2] * (num_arms - 1) + [0.5 + small_gap_delta / 2]
    core_bandit = BernoulliBandit(num_arms, horizon, means, seed)
    return core_bandit

def create_large_gap_bernoulli_bandit(num_arms, horizon, seed=None):
    big_gap_delta = 0.5
    means = [0.5 - big_gap_delta / 2] * (num_arms - 1) + [0.5 + big_gap_delta / 2]
    core_bandit = BernoulliBandit(num_arms, horizon, means, seed)
    return core_bandit

def create_high_var_gaussian_bandit(num_arms, horizon, seed=None):
    hard_variance = 3.
    rng = np.random.default_rng(seed)
    params = []
    for _ in range(num_arms):
        params.append((rng.normal(0, hard_variance), hard_variance))
    core_bandit = GaussianBandit(num_arms, horizon, params, seed)
    return core_bandit

def create_low_var_gaussian_bandit(num_arms, horizon, seed=None):
    easy_variance = 1.
    rng = np.random.default_rng(seed)
    params = []
    for _ in range(num_arms):
        params.append((rng.normal(0, easy_variance), easy_variance))
    core_bandit = GaussianBandit(num_arms, horizon, params, seed)
    return core_bandit


import math
import scipy
import numpy as np
from banditbench.tasks.mab.env import MultiArmedBandit
from banditbench.tasks.cb.env import ContextualBandit
from banditbench.tasks.cb.env import State

from typing import Union, Dict, Any

from banditbench.agents.types import MABAgent, CBAgent
from banditbench.sampling.sampler import SampleBase


class UCBAgent(MABAgent, SampleBase):
    """alpha-UCB, where alpha is the exploration bonus coefficient"""
    name: str = "UCB"

    def __init__(self, env: MultiArmedBandit, alpha: float = 0.5) -> None:
        super().__init__(env)
        self.actions = list(range(self.k_arms))
        self.alpha = alpha

        self.arms = [0] * self.k_arms  # Number of times each arm has been pulled
        self.rewards = [0] * self.k_arms  # Accumulated rewards for each arm
        self.exploration_bonuses = [0] * self.k_arms
        self.exploitation_values = [0] * self.k_arms

    def reset(self):
        self.arms = [0] * self.k_arms  # Number of times each arm has been pulled
        self.rewards = [0] * self.k_arms  # Accumulated rewards for each arm
        self.exploration_bonuses = [0] * self.k_arms
        self.exploitation_values = [0] * self.k_arms

    def calculate_arm_value(self, arm: int) -> float:
        exploration_bonus = self.calculate_exp_bonus(arm)
        exploitation_value = self.calculate_exp_value(arm)

        self.exploration_bonuses[arm] = exploration_bonus
        self.exploitation_values[arm] = exploitation_value

        return exploitation_value + exploration_bonus

    def calculate_exp_bonus(self, arm):
        # return math.sqrt(1.0 / self.arms[arm])
        return math.sqrt((self.alpha * math.log(sum(self.arms))) / self.arms[arm])

    def calculate_exp_value(self, arm):
        return self.rewards[arm] / self.arms[arm]

    def select_arm(self) -> int:
        """
        Select an arm to pull. Note that we only use self.calculate_arm_value() to select the arm.
        """
        # a hard exploration rule to always explore an arm at least once
        for arm in range(self.k_arms):
            if self.arms[arm] == 0:
                return arm  # Return an unexplored arm

        # if all arms have been explored, use UCB to select the arm
        arm_values = [self.calculate_arm_value(arm) for arm in range(self.k_arms)]
        return int(np.argmax(arm_values))

    def act(self) -> int:
        return self.select_arm()

    def update(self, action: int, reward: float, info: Dict[str, Any]) -> None:
        self.arms[action] += 1
        self.rewards[action] += reward


class GreedyAgent(UCBAgent, SampleBase):
    """
    This class shows how we can just override the calculate_arm_value() method to implement a different agent.
    """
    name: str = "Greedy"

    def __init__(self, env: MultiArmedBandit) -> None:
        super().__init__(env)

    def calculate_exp_value(self, arm: int) -> float:
        return self.rewards[arm] / self.arms[arm]


class ThompsonSamplingAgent(UCBAgent, SampleBase):
    name: str = "ThompsonSampling"

    def __init__(self, env: MultiArmedBandit, alpha_prior: float = 1.0, beta_prior: float = 1.0) -> None:
        self.alpha_prior = alpha_prior
        self.beta_prior = beta_prior
        super().__init__(env)
        self.alpha = [self.alpha_prior] * self.k_arms
        self.beta = [self.beta_prior] * self.k_arms

    def reset(self):
        self.alpha = [self.alpha_prior] * self.k_arms
        self.beta = [self.beta_prior] * self.k_arms

    def select_arm(self) -> int:
        samples = [
            scipy.stats.beta.rvs(self.alpha[arm], self.beta[arm])
            for arm in range(self.k_arms)
        ]
        return int(np.argmax(samples))

    def update(self, action: int, reward: float, info: Dict[str, Any]) -> None:
        self.alpha[action] += reward
        self.beta[action] += 1 - reward


# ======= Contextual Bandit ==========

class LinUCBAgent(CBAgent, SampleBase):
    name: str = 'LinUCB'

    def __init__(self, env: ContextualBandit, alpha: float = 0.5):
        super().__init__(env)

        self.d = env.feature_dim
        self.alpha = alpha
        self.A = [np.identity(self.d) for _ in range(self.k_arms)]
        self.b = [np.zeros((self.d, 1)) for _ in range(self.k_arms)]

    def reset(self):
        # init must be called before reset
        self.A = [np.identity(self.d) for _ in range(self.k_arms)]
        self.b = [np.zeros((self.d, 1)) for _ in range(self.k_arms)]

    def act(self, state: State) -> int:
        """Same as performing a sampling step."""
        action = self.select_action(state)
        return action

    def select_action(self, state: State) -> int:
        context = state.feature
        context = context.reshape(-1, 1)
        ucb_values = []

        ucb_exploitation_values = []
        ucb_exploration_bonuses = []

        for a in range(self.k_arms):
            A_inv = np.linalg.inv(self.A[a])
            theta = A_inv.dot(self.b[a])

            ucb = theta.T.dot(context) + self.alpha * np.sqrt(
                context.T.dot(A_inv).dot(context)
            )
            ucb_values.append(ucb[0, 0])

            ucb_exploitation_values.append(theta.T.dot(context)[0, 0])
            exp_v = self.alpha * np.sqrt(context.T.dot(A_inv).dot(context))
            ucb_exploration_bonuses.append(exp_v[0, 0])

        # tie-breaking arbitrarily
        candidate_arms = []
        highest_ucb = -1
        for arm_index in range(self.k_arms):
            # If current arm is highest than current highest_ucb
            arm_ucb = ucb_values[arm_index]
            # because we have the precision of 0.01
            if arm_ucb - highest_ucb > 0.001:
                # Set new max ucb
                highest_ucb = arm_ucb
                # Reset candidate_arms list with new entry based on current arm
                candidate_arms = [arm_index]
            # If there is a tie, append to candidate_arms
            if highest_ucb - arm_ucb <= 1e-5:
                candidate_arms.append(arm_index)

        # Choose based on candidate_arms randomly (tie breaker)
        chosen_arm_index = np.random.choice(candidate_arms)

        return int(chosen_arm_index)

    def update(self, state: State, action: int, reward: float, info: Dict[str, Any]) -> None:
        context = state.feature
        context = context.reshape(-1, 1)
        self.A[action] += context.dot(context.T)
        self.b[action] += reward * context

# add an AgentBuilder
# and a classicAgent here
# for delayed execution



from banditbench.tasks.cb.movielens import MovieLens, MovieLensVerbal
from banditbench.tasks.mab.env import BernoulliBandit, GaussianBandit, VerbalMultiArmedBandit
from banditbench.tasks.mab import (create_high_var_gaussian_bandit, create_low_var_gaussian_bandit,
                                   create_large_gap_bernoulli_bandit, create_small_gap_bernoulli_bandit)
